
from typing import Optional, List, Dict, Union
import numpy as np

from qrem.functions_qrem import ancillary_functions as anf

#ORGANISATIION:  At this moment this class is used only as a parent class to OverlappingTomographyBase.
#  In principle it owuld be good to have a it for futher generalisations (MO)
#ORGANISATION: function file does not start with capital letters (MO)

#TODO_MO (PP) This class is more about gennerating circuits, We should
# - create Circuit class. Circuits will contain circuit type definnition QCircuit =NDArray, conversions, and creation/ranndomisation functions, i/o
# - channge name to CharacterisationCircuits - as this seesm to be the merit of this class
# Further: OverlappingTomogtaphyBase should be changed into OverlappingTomography
# OverlappingTomography should get CharacterisationCircuits as an input, not as an inheritance




class OverlappingEstimationBase:
    def __init__(self,
                 number_of_qubits: int,
                 number_of_symbols: int,
                 subsets: Union[Dict[str, List[int]], List[List[int]]],
                 # maximal_circuits_amount: Optional[int] = 1500,       #THIS can potentially be restored
                 show_progress_bars: Optional[bool] = True,
                 ):

        self._number_of_symbols = number_of_symbols
        self._number_of_qubits = number_of_qubits
        self._subsets = subsets
       # self._maximal_circuits_amount = maximal_circuits_amount  #THIS can potentially be restored

        # a list that will store lists of integers that label characterisation cir
        self._circuits_list = []

        self._show_progress_bars = show_progress_bars


    # properties for restoring "inputs" of the constructor - on high level redundant

    #ALG: can we be faster when we use numpy lists here


    #------------------
    @property
    def number_of_symbols(self) -> int:
        return self._number_of_symbols

    @number_of_symbols.setter
    def number_of_symbols(self, number_of_symbols: int) -> None:
        self._number_of_symbols = number_of_symbols

    @property
    def subsets(self) -> Union[Dict[str, List[int]], List[List[int]]]:
        return self._subsets

    @subsets.setter
    def subsets(self, subsets: Union[Dict[str, List[int]], List[List[int]]]) -> None:
        self._subsets = subsets

    @property
    def circuits_list(self) -> List[List[int]]:
        return self._circuits_list

    @circuits_list.setter
    def circuits_list(self, circuits_list: List[List[int]]) -> None:
        self._circuits_list = circuits_list

    # MOquestion - what about random seed?
    def get_random_circuit(self):
        return [np.random.randint(0, self._number_of_symbols) for _ in
                range(self._number_of_qubits)]
    # Gives at output a random list of integers (from colleection of size self._number_of_symbols), where leangth correspond to number of qubits

    def get_random_circuit_with_fixed_states_of_some_qubits(self,
                                                           fixed_states: Dict[int, int]):
        """
        :param fixed_states: dictionary where each KEY is index of qubit,
                            and VALUE denotes symbols' state
        :return:
        """
# Generates fixes state on some subset of qubits, random (from   colection of size self._number_of_symbols) and appends it to internal list of circuit
# MOcomm- update documentation
        fixed_qubits = fixed_states.keys()

        circuit = []

        for qubit_index in range(self._number_of_qubits):
            if qubit_index in fixed_qubits:
                circuit.append(fixed_states[qubit_index])
            else:
                circuit.append(np.random.randint(0, self._number_of_symbols))

        return circuit

    def add_circuits(self,
                     circuits: List[List[Union[str,int]]]):
        for circuit in circuits:
            self._circuits_list.append(circuit)

    # appends a list of random  circuits of leangth number_of_circuits to internal circuit list
    def get_random_circuits_list(self,
                                 number_of_circuits: int,
                                 fixed_states: Optional[Dict[int, int]] = None):
        if fixed_states is None:
            # MOcomm - not sure about efficiency of this procedure - using numpy maybe faster
            return [self.get_random_circuit() for _ in range(number_of_circuits)]
        else:
            # MOcomm - not sure about efficiency of this procedure - using numpy maybe faster
            return [self.get_random_circuit_with_fixed_states_of_some_qubits(fixed_states) for _ in
                    range(number_of_circuits)]

    # very similar name to get_random_circuit function defined previously
    def add_random_circuits(self,
                            number_of_circuits: int,
                            fixed_states: Optional[Dict[int, int]] = None):

        self.add_circuits(self.get_random_circuits_list(number_of_circuits, fixed_states))

    # adds random circuits (encoded by integer list) to list of circuits
#MOcomm - improve documentation



"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""


from typing import List, Dict, Tuple
import numpy as np
import picos
from qrem.functions_qrem import ancillary_functions as anf

from qrem.common.printer import qprint_array


def __do_sanity_checks_stochastic_maps(positive_matrix,
                                       negative_matrix):
    dimension = positive_matrix.shape[0]
    for i in range(dimension):
        for j in range(dimension):
            el_ij_positive = positive_matrix[i, j]
            el_ij_negative = negative_matrix[i, j]

            if np.sign(el_ij_positive) == -1:
                if abs(el_ij_positive) >= 10**(-8):
                    qprint_array(positive_matrix)
                    raise ValueError(
                        f"Weird solution! Element '{(i, j)}' of positive matrix has value : {el_ij_positive}")
                positive_matrix[i, j] = abs(0.)

            if np.sign(el_ij_negative) == -1:
                if abs(el_ij_negative) >= 10**(-8):
                    qprint_array(negative_matrix)
                    raise ValueError(
                        f"Weird solution! Element '{(i, j)}' of negative matrix has value : {el_ij_negative}")
                negative_matrix[i, j] = abs(0.)

        sum_column_positive = sum(positive_matrix[:, i])
        if abs(sum_column_positive - 1) > 0.001:
            qprint_array(positive_matrix)
            raise ValueError(
                f"Weird solution! Column '{i}' of positive matrix sums to : {sum_column_positive}")
        positive_matrix[:, i] /= sum_column_positive

        sum_column_negative = sum(negative_matrix[:, i])
        if abs(sum_column_negative - 1) > 0.001:
            qprint_array(negative_matrix)
            raise ValueError(
                f"Weird solution! Column '{i}' of negative matrix sums to : {sum_column_negative}")
        negative_matrix[:, i] /= sum_column_negative

        # print(sum_column_negative,sum(negative_matrix[:,i]), sum_column_positive,sum(positive_matrix[:,i]))
        # print(positive_matrix)
        # print(negative_matrix)

    return positive_matrix, negative_matrix

def find_optimal_quasiprobability_decomposition_postprocessing(stochastic_map:np.ndarray,
                                                               ):

    dimension = stochastic_map.shape[0]


    problem = picos.Problem()
    positive, negative = picos.RealVariable(name='positive',
                                            shape=(dimension,dimension),
                                            lower=0
                                            ),\
                         picos.RealVariable(name='negative',
                                            shape=(dimension,dimension),
                                            upper=0
                                            )

    a = picos.RealVariable(name='a',
                           shape=1)
    b = 1-a

    for column_index in range(dimension):
        problem.add_constraint(picos.sum(positive[:,column_index])==a)
        problem.add_constraint(picos.sum(negative[:,column_index])==b)

    problem.add_constraint(np.linalg.inv(stochastic_map)==positive+negative)

    cost_function = a

    problem.set_objective(direction='min',
                          expression=cost_function
                          )


    problem.solve(solver='mosek')


    a_optimal = a.value
    norm_optimal = a_optimal+abs(1-a_optimal)

    positive_matrix = np.array(positive.value_as_matrix/a_optimal)
    negative_matrix = np.array(negative.value_as_matrix / (1 - a_optimal))

    positive_matrix, negative_matrix = __do_sanity_checks_stochastic_maps(positive_matrix=positive_matrix,
                                                                          negative_matrix=negative_matrix)


    return [a_optimal,1-a_optimal], [positive_matrix, negative_matrix], norm_optimal


def get_probability_distribution_from_quasiprobability(quasiprobability):

    norm = np.linalg.norm(quasiprobability,ord=1)

    return [abs(qa)/norm for qa in quasiprobability]






def get_quasiprobability_dictionary_from_noise_matrices(local_noise_matrices:Dict[Tuple[int],np.ndarray]):

    quasiprobability_dictionary = {}
    for cluster, local_noise_matrix in local_noise_matrices.items():
        quasiprobability_distribution, stochastic_maps, quasi_norm = find_optimal_quasiprobability_decomposition_postprocessing(
            local_noise_matrix)

        probability_distribution = get_probability_distribution_from_quasiprobability(
            quasiprobability=quasiprobability_distribution)

        quasiprobability_dictionary[cluster] = (probability_distribution,stochastic_maps,quasi_norm)

    return quasiprobability_dictionary







import sys
import os
import numpy as np
import pickle
from qrem.functions_qrem import ancillary_functions as anf
from qrem.common import math
from typing import List, Dict, Optional, Callable, Tuple





def Overlap(s1:str,s2:str):
    """
    Function used to create dictionary of overlapes between different Pauli states. It's needed to compute coherence indicator
    Input: s1,s2 two symbols, each corresponding to one of the Pauli states, as in the SeparableCircuitsCreator class 

    Output: value of squared scalar product between the states     
    """
    if s1==s2:
        return 1
    elif (s1=='2' and s2=='3') or (s2=='2' and s1=='3') or (s1=='4' and s2 =='5') or (s2=='4' and s1 =='5'):
        return 0
    else:
        return 0.5

def compute_indicator_normalization(dim,setting1,setting2,overlap_dic):
    """
    A function computing normalization factor of coherence indicator. As for now works for two-qubit reduced POVMs
    
    Input: 
    dim - dimension of the reduced subspace
    setting1,setting2 - strings consisting of two symbols each, corresponding to an inout Pauli eigenstate
    overlap_dic - a dictionary consisting of squared scalar product of two Pauli eigenstates   
    """
    return dim*np.sqrt((2*(1- overlap_dic[setting1[1]+setting2[1]]*overlap_dic[setting1[0]+setting2[0]])))

def compute_pauli_marginals(marginals_dictionary, subsets_list):
    """
    Dictionaries needed to perform computations

    setting dictionary - keys correspond to alphabet enocding input states
                         items correspond to unnormalized probability distributions for a given input state on subset of qubits
    normalization dictionary
                         keys correspond to alphabet encoding input states
                         items correspond to number of times that a particular input state appear in marginals

    """
    setting_dictionary = {}
    normalisation_dictionary = {}
    for i in range(2, 6):
        for j in range(2, 6):
            setting_dictionary[str(i) + str(j)] = np.array([0., 0., 0., 0.])
            normalisation_dictionary[str(i) + str(j)] = 0

    measurement_settings = marginals_dictionary.keys()
    """
    Computation of marginal probability distributions for X,Y Pauli input states  
    """
    Pauli_subset_dictionary = {}
    for subset in subsets_list:
        setting_dictionary = setting_dictionary.fromkeys(setting_dictionary, [0., 0., 0., 0.])
        normalisation_dictionary = normalisation_dictionary.fromkeys(normalisation_dictionary, 0)
        for setting in measurement_settings:
            s1 = setting[subset[0]]
            s2 = setting[subset[1]]
            if s1 != '0' and s1 != '1' and s2 != '0' and s2 != '1':
                setting_dictionary[s1 + s2] += marginals_dictionary[setting][subset]
                normalisation_dictionary[s1 + s2] += 1

        Pauli_subset_dictionary[subset] = [setting_dictionary, normalisation_dictionary]
    return Pauli_subset_dictionary



def compute_coherence_indicator(marginals_dictionary, subset_list,settings_list=['2','3','4','5']):
    """
    Computation of coherence indicator 

    TVD between probability distributions generated by different input X,Y Pauli eigenstates of two qubits
    """
    pauli_subset_dictionary = compute_pauli_marginals(marginals_dictionary,subset_list)

    """
    Creation of overlap dicitionary for the Pauli case
    """
   

    overlap_dictionary={}
    settings_list_coherence_indicator = []
    for i in settings_list:
        for j in settings_list:
            overlap_dictionary[i+j] = Overlap(i,j)
            settings_list_coherence_indicator.append(i+j)

   

    # dictionary storing values of coherence indicator
    indicator_dic = {}

  

    for keys, elements in pauli_subset_dictionary.items():
        tvd_value = []
        tvd_settings = []

        for i in range(len(settings_list_coherence_indicator)):
            for j in range(i + 1, len(settings_list_coherence_indicator)):
                s1 = settings_list_coherence_indicator[i]
                s2 = settings_list_coherence_indicator[j]
                indicator = math.compute_TVD(elements[0][s1] / elements[1][s1], elements[0][s2] / elements[1][s2])
                indicator = indicator / (compute_indicator_normalization(2, s1, s2, overlap_dictionary))
                tvd_value.append(indicator)
                tvd_settings.append((settings_list_coherence_indicator[i], settings_list_coherence_indicator[j]))
        indicator_dic[keys] = [tvd_value, tvd_settings]

    return  indicator_dic

if __name__ == "__main__":
    testcase1 = math.compute_TVD(np.array([1,0]),np.array([0,1]))
    print(testcase1)
    


"""
@authors: Filip Maciejewski, Oskar Słowik, Tomek Rybotycki
@contact: filip.b.maciejewski@gmail.com
"""

from base_classes.heuristic_clustering_algorithm_base import HeuristicClusteringAlgorithmBase
from qrem.functions_qrem.functions_noise_model_heuristic_help import get_avg_large_corr, suggest_alpha_via_scan

class HeuristicClusteringAlgorithmAuto(HeuristicClusteringAlgorithmBase):

    def __init__(self, **kwargs):
        super(HeuristicClusteringAlgorithmAuto, self).__init__(**kwargs)

    def suggest_alpha_heuristic(self):
        if self.correlations_table is None:
            raise ValueError("correlations_table not specified.")
        if self.C_maxsize is None:
            raise ValueError("C_maxsize not specified.")
        return 0.5 * get_avg_large_corr(self.correlations_table, self.C_maxsize)

    def suggest_alpha_scan(self, values_to_return=None):
        return suggest_alpha_via_scan(self, values_to_return)



# -*- coding: utf-8 -*-
"""
The module converts a noncommutative optimization problem provided in SymPy
format to an SDPA semidefinite programming problem.

Created on Sun May 26 15:06:17 2013

@author: Peter Wittek
"""
from __future__ import division, print_function
import os
from sympy import S, zeros
from sympy.physics.quantum.dagger import Dagger
import tempfile
from .nc_utils import apply_substitutions, is_number_type
from .sdp_relaxation import SdpRelaxation
from .sdpa_utils import write_to_sdpa


class SteeringHierarchy(SdpRelaxation):

    """Class for obtaining a step in the steering hierarchy.

    :param variables: Commutative or noncommutative, Hermitian or nonhermiatian
                      variables.
    :type variables: list of :class:`sympy.physics.quantum.operator.Operator`
                     or
                     :class:`sympy.physics.quantum.operator.HermitianOperator`
                     or a list of list.
    :param verbose: Optional parameter for level of verbosity:

                       * 0: quiet
                       * 1: verbose
                       * 2: debug level

    :type verbose: int.
    :param matrix_var_dim: Optional parameter to specify the size of matrix
                           variable blocks
    :type matrix_var_dim: int.
    :param mark_conjugate: Use this optional parameter to generate a symbolic
                           representation of the steering hierarchy for export.
    :type mark_conjugate: bool.

    Attributes:
      - `monomial_sets`: The monomial sets that generate the moment matrix blocks.

      - `monomial_index`: Dictionary that maps monomials to SDP variables.

      - `constraints`: The complete set of constraints after preprocesssing.

      - `primal`: The primal optimal value.

      - `dual`: The dual optimal value.

      - `x_mat`: The primal solution matrix.

      - `y_mat`: The dual solution matrix.

      - `solution_time`: The amount of time taken to solve the relaxation.

      - `status`: The solution status of the relaxation.

    """
    hierarchy_types = ["npa", "npa_chordal", "moroder"]

    def __init__(self, variables, verbose=0, matrix_var_dim=None,
                 mark_conjugate=False, parallel=False):
        """Constructor for the class.
        """
        if matrix_var_dim is None and not mark_conjugate:
            raise Exception("Invalid steering hierarchy requested!")
        super(SteeringHierarchy, self).__init__(variables, verbose=verbose,
                                                parallel=parallel)
        if matrix_var_dim is not None:
            self.complex_matrix = True
            self.normalized = False
        self.matrix_var_dim = matrix_var_dim
        self.mark_conjugate = mark_conjugate

    def _process_monomial(self, monomial, n_vars):
        """Process a single monomial when building the moment matrix.
        """
        coeff, monomial = monomial.as_coeff_Mul()
        k = 0
        # Have we seen this monomial before?
        conjugate = False
        try:
            # If yes, then we improve sparsity by reusing the
            # previous variable to denote this entry in the matrix
            k = self.monomial_index[monomial]
        except KeyError:
            # An extra round of substitutions is granted on the conjugate of
            # the monomial if all the variables are Hermitian
            daggered_monomial = \
                apply_substitutions(Dagger(monomial), self.substitutions,
                                    self.pure_substitution_rules)
            try:
                k = self.monomial_index[daggered_monomial]
                conjugate = True
            except KeyError:
                # Otherwise we define a new entry in the associated
                # array recording the monomials, and add an entry in
                # the moment matrix
                k = n_vars + 1
                self.monomial_index[monomial] = k
        if conjugate:
            k = -k
        return k, coeff

    def _push_monomial(self, monomial, n_vars, row_offset, rowA, columnA, N,
                       rowB, columnB, lenB, prevent_substitutions=False):
        monomial = apply_substitutions(monomial, self.substitutions,
                                       self.pure_substitution_rules)
        if is_number_type(monomial):
            if rowA == 0 and columnA == 0 and rowB == 0 and columnB == 0 and \
                    monomial == 1.0 and not self.normalized:
                if self.matrix_var_dim is None:
                    n_vars += 1
                    self.F[row_offset + rowA * N*lenB + rowB * N +
                           columnA * lenB + columnB, n_vars] = 1
                else:
                    n_vars = self.__add_matrix_variable(row_offset, rowA,
                                                        columnA, N, rowB,
                                                        columnB, lenB,
                                                        n_vars + 1, False, 1)
            else:
                self.F[row_offset + rowA * N*lenB + rowB * N +
                       columnA * lenB + columnB, 0] = monomial
        elif monomial.is_Add:
            for element in monomial.as_ordered_terms():
                n_vars = self._push_monomial(element, n_vars, row_offset,
                                             rowA, columnA, N,
                                             rowB, columnB, lenB)
        elif monomial != 0:
            k, coeff = self._process_monomial(monomial, n_vars)
            # We push the entry to the moment matrix
            if self.matrix_var_dim is None:
                if self.mark_conjugate:
                    if k < 0:
                        coeff = -coeff
                        k = -k
                self.F[row_offset + rowA * N*lenB + rowB * N +
                       columnA * lenB + columnB, k] = coeff
            else:
                conjugate = False
                if k < 0:
                    conjugate = True
                    k = -k
                k = self.__add_matrix_variable(row_offset, rowA, columnA, N,
                                               rowB, columnB, lenB, k,
                                               conjugate, coeff)
            if k > n_vars:
                n_vars = k
        return n_vars

    def __add_matrix_variable(self, row_offset, rowA, columnA, N, rowB,
                              columnB, lenB, k, conjugate, coeff):
        if conjugate:
            imag = -1j
        else:
            imag = 1j
        for sub_row in range(self.matrix_var_dim):
            for sub_column in range(self.matrix_var_dim):
                if conjugate:
                    r, c = sub_column, sub_row
                else:
                    r, c = sub_row, sub_column
                block_row_index = rowA*self.matrix_var_dim + r
                block_column_index = columnA*self.matrix_var_dim + c
                if block_row_index > block_column_index:
                    k += 1
                    continue
                if block_row_index == block_column_index:
                    imag_zero = 0
                else:
                    imag_zero = 1
                value = coeff*(1+imag*imag_zero)
                self.F[row_offset +
                       block_row_index*self.matrix_var_dim*N*lenB +
                       rowB*self.matrix_var_dim*N + block_column_index*lenB +
                       self.matrix_var_dim*columnB, k] = value
                k += 1
        k -= 1
        return k

    def __get_trace_facvar(self, polynomial):
        """Return dense vector representation of a polynomial. This function is
        nearly identical to __push_facvar_sparse, but instead of pushing
        sparse entries to the constraint matrices, it returns a dense
        vector.
        """
        facvar = [0] * (self.n_vars + 1)
        F = {}
        for i in range(self.matrix_var_dim):
            for j in range(self.matrix_var_dim):
                for key, value in \
                        polynomial[i, j].as_coefficients_dict().items():
                    skey = apply_substitutions(key, self.substitutions,
                                               self.pure_substitution_rules)
                    try:
                        Fk = F[skey]
                    except KeyError:
                        Fk = zeros(self.matrix_var_dim, self.matrix_var_dim)
                    Fk[i, j] += value
                    F[skey] = Fk
        # This is the tracing part
        for key, Fk in F.items():
            if key == S.One:
                k = 1
            else:
                k = self.monomial_index[key]
            for i in range(self.matrix_var_dim):
                for j in range(self.matrix_var_dim):
                    sym_matrix = zeros(self.matrix_var_dim,
                                       self.matrix_var_dim)
                    sym_matrix[i, j] = 1
                    facvar[k+i*self.matrix_var_dim+j] = (sym_matrix*Fk).trace()
        facvar = [float(f) for f in facvar]
        return facvar

    def set_objective(self, objective, extraobjexpr=None):
        """Set or change the objective function of the polynomial optimization
        problem.

        :param objective: Describes the objective function.
        :type objective: :class:`sympy.core.expr.Expr`
        :param extraobjexpr: Optional parameter of a string expression of a
                             linear combination of moment matrix elements to be
                             included in the objective function
        :type extraobjexpr: str.
        """
        if objective is not None and self.matrix_var_dim is not None:
            facvar = self.__get_trace_facvar(objective)
            self.obj_facvar = facvar[1:]
            self.constant_term = facvar[0]
            if self.verbose > 0 and facvar[0] != 0:
                print("Warning: The objective function has a non-zero %s "
                      "constant term. It is not included in the SDP objective."
                      % facvar[0])
        else:
            super(SteeringHierarchy, self).\
              set_objective(objective, extraobjexpr=extraobjexpr)

    def _calculate_block_structure(self, inequalities, equalities,
                                   momentinequalities, momentequalities,
                                   extramomentmatrix, removeequalities,
                                   block_struct=None):
        """Calculates the block_struct array for the output file.
        """
        super(SteeringHierarchy, self).\
          _calculate_block_structure(inequalities, equalities,
                                     momentinequalities, momentequalities,
                                     extramomentmatrix, removeequalities)
        if self.matrix_var_dim is not None:
            self.block_struct = [self.matrix_var_dim*bs
                                 for bs in self.block_struct]

    def _estimate_n_vars(self):
        if self.matrix_var_dim is None:
            super(SteeringHierarchy, self)._estimate_n_vars()
        else:
            self.n_vars = 0
            for monomials in self.monomial_sets:
                n_monomials = len(monomials)
                self.n_vars += int(n_monomials * (n_monomials + 1) / 2) *\
                               self.matrix_var_dim**2

    def write_to_file(self, filename, filetype=None):
        """Write the relaxation to a file.

        :param filename: The name of the file to write to. The type can be
                         autodetected from the extension: .dat-s for SDPA,
                         .task for mosek, .csv for human readable format, or
                         .txt for a symbolic export
        :type filename: str.
        :param filetype: Optional parameter to define the filetype. It can be
                         "sdpa" for SDPA , "mosek" for Mosek, "csv" for
                         human readable format, or "txt" for a symbolic export.
        :type filetype: str.
        """
        if filetype == "txt" and not filename.endswith(".txt"):
            raise Exception("TXT files must have .txt extension!")
        elif filetype is None and filename.endswith(".txt"):
            filetype = "txt"
        else:
            return super(SteeringHierarchy, self).write_to_file(filename,
                                                                filetype=filetype)
        tempfile_ = tempfile.NamedTemporaryFile()
        tmp_filename = tempfile_.name
        tempfile_.close()
        tmp_dats_filename = tmp_filename + ".dat-s"
        write_to_sdpa(self, tmp_dats_filename)
        f = open(tmp_dats_filename, 'r')
        f.readline();f.readline();f.readline()
        blocks = ((f.readline().strip().split(" = ")[0])[1:-1]).split(", ")
        block_offset, matrix_size = [0], 0
        for block in blocks:
            matrix_size += abs(int(block))
            block_offset.append(matrix_size)
        f.readline()
        matrix = [[0 for _ in range(matrix_size)] for _ in range(matrix_size)]
        for line in f:
            entry = line.strip().split("\t")
            var, block = int(entry[0]), int(entry[1])-1
            row, column = int(entry[2]) - 1, int(entry[3]) - 1
            value = float(entry[4])
            offset = block_offset[block]
            matrix[offset+row][offset+column] = int(value*var)
            matrix[offset+column][offset+row] = int(value*var)
        f.close()
        f = open(filename, 'w')
        for matrix_line in matrix:
            f.write(str(matrix_line).replace('[', '').replace(']', '') + '\n')
        f.close()
        os.remove(tmp_dats_filename)



# -*- coding: utf-8 -*-
"""
This file contains helper functions to work with noncommutative polynomials
and Hamiltonians.

Created on Thu May  2 16:03:05 2013

@author: Peter Wittek
"""
from __future__ import division, print_function
from sympy import adjoint, conjugate, S, Symbol, Pow, Number, expand, I
from sympy.physics.quantum import HermitianOperator, Operator
from scipy.sparse import lil_matrix


def flatten(lol):
    """Flatten a list of lists to a list.

    :param lol: A list of lists in arbitrary depth.
    :type lol: list of list.

    :returns: flat list of elements.
    """
    new_list = []
    for element in lol:
        if element is None:
            continue
        elif not isinstance(element, list) and not isinstance(element, tuple):
            new_list.append(element)
        elif len(element) > 0:
            new_list.extend(flatten(element))
    return new_list


def simplify_polynomial(polynomial, monomial_substitutions):
    """Simplify a polynomial for uniform handling later.
    """
    if isinstance(polynomial, (int, float, complex)):
        return polynomial
    polynomial = (1.0 * polynomial).expand(mul=True,
                                           multinomial=True)
    if is_number_type(polynomial):
        return polynomial
    if polynomial.is_Mul:
        elements = [polynomial]
    else:
        elements = polynomial.as_coeff_mul()[1][0].as_coeff_add()[1]
    new_polynomial = 0
    # Identify its constituent monomials
    for element in elements:
        monomial, coeff = separate_scalar_factor(element)
        monomial = apply_substitutions(monomial, monomial_substitutions)
        new_polynomial += coeff * monomial
    return new_polynomial


def is_pure_substitution_rule(lhs, rhs):
    if is_number_type(rhs):
        return True
    if rhs.is_Mul:
        elements = [rhs]
    else:
        elements = rhs.as_coeff_mul()[1][0].as_coeff_add()[1]
    for element in elements:
        monomial, _ = separate_scalar_factor(element)
        for atom in monomial.atoms():
            if atom.is_Number:
                continue
            if not lhs.has(atom):
                return False
    return True


def remove_scalar_factor(monomial):
    """Return monomial without constant factor.
    """
    monomial, _ = separate_scalar_factor(monomial)
    return monomial


def __separate_scalar_factor(monomial):
    """Separate the constant factor from a monomial.
    """
    scalar_factor = 1
    if is_number_type(monomial):
        return S.One, monomial
    if monomial == 0:
        return S.One, 0
    comm_factors, _ = split_commutative_parts(monomial)
    if len(comm_factors) > 0:
        if isinstance(comm_factors[0], Number):
            scalar_factor = comm_factors[0]
    if scalar_factor != 1:
        return monomial / scalar_factor, scalar_factor
    else:
        return monomial, scalar_factor


def get_support(variables, polynomial):
    """Gets the support of a polynomial.
    """
    support = []
    if is_number_type(polynomial):
        support.append([0] * len(variables))
        return support
    for monomial in polynomial.expand().as_coefficients_dict():
        tmp_support = [0] * len(variables)
        mon, _ = __separate_scalar_factor(monomial)
        symbolic_support = flatten(split_commutative_parts(mon))
        for s in symbolic_support:
            if isinstance(s, Pow):
                base = s.base
                if is_adjoint(base):
                    base = base.adjoint()
                tmp_support[variables.index(base)] = s.exp
            elif is_adjoint(s):
                tmp_support[variables.index(s.adjoint())] = 1
            elif isinstance(s, (Operator, Symbol)):
                tmp_support[variables.index(s)] = 1
        support.append(tmp_support)
    return support


def get_support_variables(polynomial):
    """Gets the support of a polynomial.
    """
    support = []
    if is_number_type(polynomial):
        return support
    for monomial in polynomial.expand().as_coefficients_dict():
        mon, _ = __separate_scalar_factor(monomial)
        symbolic_support = flatten(split_commutative_parts(mon))
        for s in symbolic_support:
            if isinstance(s, Pow):
                base = s.base
                if is_adjoint(base):
                    base = base.adjoint()
                support.append(base)
            elif is_adjoint(s):
                support.append(s.adjoint())
            elif isinstance(s, Operator):
                support.append(s)
    return support


def separate_scalar_factor(element):
    """Construct a monomial with the coefficient separated
    from an element in a polynomial.
    """
    coeff = 1.0
    monomial = S.One
    if isinstance(element, (int, float, complex)):
        coeff *= element
        return monomial, coeff
    for var in element.as_coeff_mul()[1]:
        if not (var.is_Number or var.is_imaginary):
            monomial = monomial * var
        else:
            if var.is_Number:
                coeff = float(var)
            # If not, then it is imaginary
            else:
                coeff = 1j * coeff
    coeff = float(element.as_coeff_mul()[0]) * coeff
    return monomial, coeff


def count_ncmonomials(monomials, degree):
    """Given a list of monomials, it counts those that have a certain degree,
    or less. The function is useful when certain monomials were eliminated
    from the basis.

    :param variables: The noncommutative variables making up the monomials
    :param monomials: List of monomials (the monomial basis).
    :param degree:  Maximum degree to count.

    :returns: The count of appropriate monomials.
    """
    ncmoncount = 0
    for monomial in monomials:
        if ncdegree(monomial) <= degree:
            ncmoncount += 1
        else:
            break
    return ncmoncount


def apply_substitutions(monomial, monomial_substitutions, pure=False):
    """Helper function to remove monomials from the basis."""
    if is_number_type(monomial):
        return monomial
    original_monomial = monomial
    changed = True
    if not pure:
        substitutions = monomial_substitutions
    else:
        substitutions = {}
        for lhs, rhs in monomial_substitutions.items():
            irrelevant = False
            for atom in lhs.atoms():
                if atom.is_Number:
                    continue
                if not monomial.has(atom):
                    irrelevant = True
                    break
            if not irrelevant:
                substitutions[lhs] = rhs
    while changed:
        for lhs, rhs in substitutions.items():
            monomial = fast_substitute(monomial, lhs, rhs)
        if original_monomial == monomial:
            changed = False
        original_monomial = monomial
    return monomial


def fast_substitute(monomial, old_sub, new_sub):
    """Experimental fast substitution routine that considers only restricted
    cases of noncommutative algebras. In rare cases, it fails to find a
    substitution. Use it with proper testing.

    :param monomial: The monomial with parts need to be substituted.
    :param old_sub: The part to be replaced.
    :param new_sub: The replacement.
    """
    # If monomial is number -> nothing to do
    if is_number_type(monomial):
        return monomial
    # If polynomial then apply to each term in sum
    if monomial.is_Add:
        return sum([fast_substitute(element, old_sub, new_sub) for element in
                    monomial.as_ordered_terms()])

    # Split the monomial into its factors (commutative separated)
    comm_factors, ncomm_factors = split_commutative_parts(monomial)
    # Split the part to be substituted into factors
    old_comm_factors, old_ncomm_factors = split_commutative_parts(old_sub)
    # This is a temporary hack
    # If the new_part is not a number then get its commutative parts
    if not isinstance(new_sub, (int, float, complex)):
        new_comm_factors, _ = split_commutative_parts(new_sub)
    else:
        new_comm_factors = [new_sub]
    comm_monomial = 1
    is_constant_term = False
    # If we have commutative variables then deal with them here
    if comm_factors != ():
        if len(comm_factors) == 1 and is_number_type(comm_factors[0]):
            is_constant_term = True
            comm_monomial = comm_factors[0]
        else:
            for comm_factor in comm_factors:
                comm_monomial *= comm_factor
            if old_comm_factors != ():
                comm_old_sub = 1
                for comm_factor in old_comm_factors:
                    comm_old_sub *= comm_factor
                comm_new_sub = 1
                for comm_factor in new_comm_factors:
                    comm_new_sub *= comm_factor
                # Dummy heuristic to get around retarded SymPy bug
                if isinstance(comm_old_sub, Pow):
                    # In this case, we are in trouble
                    old_base = comm_old_sub.base
                    if comm_monomial.has(old_base):
                        old_degree = comm_old_sub.exp
                        new_monomial = 1
                        match = False
                        for factor in comm_monomial.as_ordered_factors():
                            if factor.has(old_base):
                                if isinstance(factor, Pow):
                                    degree = factor.exp
                                    if degree >= old_degree:
                                        match = True
                                        new_monomial *= \
                                            old_base**(degree-old_degree) * \
                                            comm_new_sub

                                else:
                                    new_monomial *= factor
                            else:
                                new_monomial *= factor
                        if match:
                            comm_monomial = new_monomial
                else:
                    comm_monomial = comm_monomial.subs(comm_old_sub,
                                                       comm_new_sub)
    if ncomm_factors == () or old_ncomm_factors == ():
        return comm_monomial

    # old_factors = old_sub.as_ordered_factors()
    # factors = monomial.as_ordered_factors()
    # Now dealing with the non_comm factors
    new_var_list = []
    new_monomial = 1
    left_remainder = 1
    right_remainder = 1
    # ncomm_factors -- factors of the monomial
    # old_ncomm_factors -- factors of the part to be substituted
    # We move through the monomial, checking that the substituted part
    # lines up with some part of the monomial (if it does then we substitute)
    for i in range(len(ncomm_factors) - len(old_ncomm_factors) + 1):
        for j, old_ncomm_factor in enumerate(old_ncomm_factors):
            ncomm_factor = ncomm_factors[i + j]
            # If monomial factor is Symbol AND
            # [sub_part is not symbol or the same symbol then move to next mono_part]
            if isinstance(ncomm_factor, Symbol) and \
                (isinstance(old_ncomm_factor, Operator) or
                 (isinstance(old_ncomm_factor, Symbol) and
                  ncomm_factor != old_ncomm_factor)):
                left_remainder, right_remainder = 1, 1
                break
            # If monomial fact is Operator AND
            # [sub_part is not operator or not same operator then move to next mono_part]
            if isinstance(ncomm_factor, Operator) and \
                    ((isinstance(old_ncomm_factor, Operator) and
                      ncomm_factor != old_ncomm_factor) or
                     isinstance(old_ncomm_factor, Pow)):
                left_remainder, right_remainder = 1, 1
                break
            # If mono_part is daggered:
            #   if sub is not daggered or doesnt equal mono_part then move to next mono_part
            if is_adjoint(ncomm_factor):
                if not is_adjoint(old_ncomm_factor) or \
                         ncomm_factor != old_ncomm_factor:
                    left_remainder, right_remainder = 1, 1
                    break
            # If mono_part is not adjoint
            else:
                # if mono_part is not a pow  but sub_point is adjoint then move to next mono_part
                if not isinstance(ncomm_factor, Pow):
                    if is_adjoint(old_ncomm_factor):
                        left_remainder, right_remainder = 1, 1
                        break
                # if mono_part is a pow
                else:
                    # if sub_part is also a pow  then split the base and exponent
                    if isinstance(old_ncomm_factor, Pow):
                        old_base = old_ncomm_factor.base
                        old_degree = old_ncomm_factor.exp
                    else:
                        old_base = old_ncomm_factor
                        old_degree = 1
                    # If the bases don't match then move to next mono part
                    if old_base != ncomm_factor.base:
                        left_remainder, right_remainder = 1, 1
                        break
                    # If the exponent of sub part is too large then move to next mono_part
                    if old_degree > ncomm_factor.exp:
                        left_remainder, right_remainder = 1, 1
                        break
                    # If the exponent of sub_part is smaller than the monomial part
                    if old_degree < ncomm_factor.exp:
                        # if we are not at the last factor of sub_part
                        if j != len(old_ncomm_factors) - 1:
                            # neither last or first factor then sub doesn't fit and break
                            if j != 0:
                                left_remainder, right_remainder = 1, 1
                                break
                            # We must now be the left-most factor of sub
                            # So we can potentially sub in with some remainer to the left
                            else:
                                left_remainder = old_base ** (
                                    ncomm_factor.exp - old_degree)
                        # if we are the last factor then we can sub in with remainer to the right
                        else:
                            right_remainder = old_base ** (
                                ncomm_factor.exp - old_degree)
        # This else is executed after the inner-loop has been exhausted (no break)
        # This should only happen if we find a match for the substitution
        else:
            new_monomial = 1
            # All parts of mono up to the substituted bit
            for var in new_var_list:
                new_monomial *= var
            new_monomial *= left_remainder * new_sub * right_remainder
            for j in range(i + len(old_ncomm_factors), len(ncomm_factors)):
                new_monomial *= ncomm_factors[j]
            new_monomial *= comm_monomial
            break
        # Each time we broke out of the inner loop we found another part
        # of the monomial that will not be substituted. Add it to the new_var_list
        new_var_list.append(ncomm_factors[i])
    # We broke out of the outer loop
    else:
        if not is_constant_term and comm_factors != ():
            new_monomial = comm_monomial
            for factor in ncomm_factors:
                new_monomial *= factor
        else:
            return monomial
    if not isinstance(new_sub, (float, int, complex)) and new_sub.is_Add:
        return expand(new_monomial)
    else:
        return new_monomial


def generate_variables(name, n_vars=1, hermitian=None, commutative=True):
    """Generates a number of commutative or noncommutative variables

    :param name: The prefix in the symbolic representation of the noncommuting
                 variables. This will be suffixed by a number from 0 to
                 n_vars-1 if n_vars > 1.
    :type name: str.
    :param n_vars: The number of variables.
    :type n_vars: int.
    :param hermitian: Optional parameter to request Hermitian variables .
    :type hermitian: bool.
    :param commutative: Optional parameter to request commutative variables.
                        Commutative variables are Hermitian by default.
    :type commutative: bool.

    :returns: list of :class:`sympy.physics.quantum.operator.Operator` or
              :class:`sympy.physics.quantum.operator.HermitianOperator`
              variables or `sympy.Symbol`

    :Example:

    >>> generate_variables('y', 2, commutative=True)
    ￼[y0, y1]
    """

    variables = []
    for i in range(n_vars):
        if n_vars > 1:
            var_name = '%s%s' % (name, i)
        else:
            var_name = '%s' % name
        if commutative:
            if hermitian is None or hermitian:
                variables.append(Symbol(var_name, real=True))
            else:
                variables.append(Symbol(var_name, complex=True))
        elif hermitian is not None and hermitian:
            variables.append(HermitianOperator(var_name))
        else:
            variables.append(Operator(var_name))
    return variables


def generate_operators(name, n_vars=1, hermitian=None, commutative=False):
    """Generates a number of commutative or noncommutative operators

    :param name: The prefix in the symbolic representation of the noncommuting
                 variables. This will be suffixed by a number from 0 to
                 n_vars-1 if n_vars > 1.
    :type name: str.
    :param n_vars: The number of variables.
    :type n_vars: int.
    :param hermitian: Optional parameter to request Hermitian variables .
    :type hermitian: bool.
    :param commutative: Optional parameter to request commutative variables.
                        Commutative variables are Hermitian by default.
    :type commutative: bool.

    :returns: list of :class:`sympy.physics.quantum.operator.Operator` or
              :class:`sympy.physics.quantum.operator.HermitianOperator`
              variables

    :Example:

    >>> generate_variables('y', 2, commutative=True)
    ￼[y0, y1]
    """

    variables = []
    for i in range(n_vars):
        if n_vars > 1:
            var_name = '%s%s' % (name, i)
        else:
            var_name = '%s' % name
        if hermitian is not None and hermitian:
            variables.append(HermitianOperator(var_name))
        else:
            variables.append(Operator(var_name))
        variables[-1].is_commutative = commutative
    return variables


def get_monomials(variables, degree):
    """Generates all noncommutative monomials up to a degree

    :param variables: The noncommutative variables to generate monomials from
    :type variables: list of :class:`sympy.physics.quantum.operator.Operator`
                     or
                     :class:`sympy.physics.quantum.operator.HermitianOperator`.
    :param degree: The maximum degree.
    :type degree: int.

    :returns: list of monomials.
    """
    if degree == -1:
        return []
    if not variables:
        return [S.One]
    else:
        _variables = variables[:]
        _variables.insert(0, 1)
        ncmonomials = [S.One]
        ncmonomials.extend(var for var in variables)
        for var in variables:
            if not is_hermitian(var):
                ncmonomials.append(var.adjoint())
        for _ in range(1, degree):
            temp = []
            for var in _variables:
                for new_var in ncmonomials:
                    temp.append(var * new_var)
                    if var != 1 and not is_hermitian(var):
                        temp.append(var.adjoint() * new_var)
            ncmonomials = unique(temp[:])
        return ncmonomials


def ncdegree(polynomial):
    """Returns the degree of a noncommutative polynomial.

    :param polynomial: Polynomial of noncommutive variables.
    :type polynomial: :class:`sympy.core.expr.Expr`.

    :returns: int -- the degree of the polynomial.
    """
    degree = 0
    if is_number_type(polynomial):
        return degree
    polynomial = polynomial.expand()
    for monomial in polynomial.as_coefficients_dict():
        subdegree = 0
        for variable in monomial.as_coeff_mul()[1]:
            if isinstance(variable, Pow):
                subdegree += variable.exp
            elif not isinstance(variable, Number) and variable != I:
                subdegree += 1
        if subdegree > degree:
            degree = subdegree
    return degree


def iscomplex(polynomial):
    """Returns whether the polynomial has complex coefficients

    :param polynomial: Polynomial of noncommutive variables.
    :type polynomial: :class:`sympy.core.expr.Expr`.

    :returns: bool -- whether there is a complex coefficient.
    """
    if isinstance(polynomial, (int, float)):
        return False
    if isinstance(polynomial, complex):
        return True
    polynomial = polynomial.expand()
    for monomial in polynomial.as_coefficients_dict():
        for variable in monomial.as_coeff_mul()[1]:
            if isinstance(variable, complex) or variable == I:
                return True
    return False


def get_all_monomials(variables, extramonomials, substitutions, degree,
                      removesubstitutions=True):
    """Return the monomials of a certain degree.
    """
    monomials = get_monomials(variables, degree)
    if extramonomials is not None:
        monomials.extend(extramonomials)
    if removesubstitutions and substitutions is not None:
        monomials = [monomial for monomial in monomials if monomial not
                     in substitutions]
        monomials = [remove_scalar_factor(apply_substitutions(monomial,
                                                              substitutions))
                     for monomial in monomials]
    monomials = unique(monomials)
    return monomials


def pick_monomials_up_to_degree(monomials, degree):
    """Collect monomials up to a given degree.
    """
    ordered_monomials = []
    if degree >= 0:
        ordered_monomials.append(S.One)
    for deg in range(1, degree + 1):
        ordered_monomials.extend(pick_monomials_of_degree(monomials, deg))
    return ordered_monomials


def pick_monomials_of_degree(monomials, degree):
    """Collect all monomials up of a given degree.
    """
    selected_monomials = []
    for monomial in monomials:
        if ncdegree(monomial) == degree:
            selected_monomials.append(monomial)
    return selected_monomials


def convert_monomial_to_string(monomial):
    monomial_str = ('%s' % monomial)
    monomial_str = monomial_str.replace('Dagger(', '')
    monomial_str = monomial_str.replace(')', 'T')
    monomial_str = monomial_str.replace('**', '^')
    return monomial_str


def save_monomial_index(filename, monomial_index):
    """Save a monomial dictionary for debugging purposes.

    :param filename: The name of the file to save to.
    :type filename: str.
    :param monomial_index: The monomial index of the SDP relaxation.
    :type monomial_index: dict of :class:`sympy.core.expr.Expr`.

    """
    monomial_translation = [''] * (len(monomial_index) + 1)
    for key, k in monomial_index.items():
        monomial_translation[k] = convert_monomial_to_string(key)
    file_ = open(filename, 'w')
    for k in range(len(monomial_translation)):
        file_.write('%s %s\n' % (k, monomial_translation[k]))
    file_.close()


def unique(seq):
    """Helper function to include only unique monomials in a basis."""
    seen = {}
    result = []
    for item in seq:
        marker = item
        if marker in seen:
            continue
        seen[marker] = 1
        result.append(item)
    return result


def build_permutation_matrix(permutation):
    """Build a permutation matrix for a permutation.
    """
    matrix = lil_matrix((len(permutation), len(permutation)))
    column = 0
    for row in permutation:
        matrix[row, column] = 1
        column += 1
    return matrix


def convert_relational(relational):
    """Convert all inequalities to >=0 form.
    """
    rel = relational.rel_op
    if rel in ['==', '>=', '>']:
        return relational.lhs-relational.rhs
    elif rel in ['<=', '<']:
        return relational.rhs-relational.lhs
    else:
        raise Exception("The relational operation ' + rel + ' is not "
                        "implemented!")


def find_variable_set(variable_sets, polynomial):
    if not isinstance(variable_sets[0], list):
        return 0
    support = set(get_support_variables(polynomial))
    for i, variable_set in enumerate(variable_sets):
        if len(support-set(variable_set)) == 0:
            return i
    return -1


def moment_of_entry(pos, monomials, ineq, substitutions):
    row = pos[0]
    column = pos[1]
    if isinstance(ineq, str):
        return row, column, ineq
    else:
        return row, column, simplify_polynomial(monomials[row].adjoint() *
                                                ineq * monomials[column],
                                                substitutions)


def assemble_monomial_and_do_substitutions(arg, monomialsA, monomialsB, ppt,
                                           substitutions,
                                           pure_substitution_rules):
    rowA = arg[0]
    columnA = arg[1]
    rowB = arg[2]
    columnB = arg[3]
    if (not ppt) or (columnB >= rowB):
        monomial = monomialsA[rowA].adjoint() * monomialsA[columnA] * \
            monomialsB[rowB].adjoint() * monomialsB[columnB]
    else:
        monomial = monomialsA[rowA].adjoint() * monomialsA[columnA] * \
            monomialsB[columnB].adjoint() * monomialsB[rowB]
        # Apply the substitutions if any
    monomial = apply_substitutions(monomial, substitutions,
                                   pure_substitution_rules)

    return columnA, rowB, columnB, monomial


def split_commutative_parts(e):
    if e.is_Mul:
        args = e.args
    else:
        args = (e,)
    for i, m in enumerate(args):
        if not m.is_commutative:
            break
    else:
        i += 1
    return args[:i], args[i:]


def is_number_type(exp):
    return isinstance(exp, (int, float, complex)) or exp.is_number


def is_adjoint(exp):
    return isinstance(exp, (adjoint, conjugate))


def is_hermitian(exp):
    return exp.is_hermitian or (exp.is_hermitian is None and exp.is_complex)


def flip_sign(exp):
    if isinstance(exp, str):
        if not exp.startswith("+") and not exp.startswith("+"):
            tmp = "+" + exp
        else:
            tmp = exp
        tmp = tmp.replace("+", "p")
        tmp = tmp.replace("-", "+")
        tmp = tmp.replace("p", "-")
        return tmp
    else:
        return -exp



import pickle 

from tqdm import tqdm
import numpy as np

import qrem.common.providers.ibmutils.data_converters




from qrem.cn import simulation as cnsimulation 
from qrem.usecases.mitigation import create_mitigation_data_KKM as mitigation_routines 
from qrem.functions_qrem import functions_benchmarks as fun_ben
from qrem.functions_qrem import ancillary_functions as anf
from qrem.functions_qrem import functions_data_analysis as fdt
from qrem.types import cn_noise_model 
from datetime import date
import statistics






        




if __name__ == '__main__':
 




    #specify n#
    number_of_qubits = 100
    #version_name = 'v4'
    


    #################################################################################
    # Realization: custom_n (n=1,2,3...10) fixed noise model tested for all versions v1,v2,v3,v4,v4_temp (temp =1, temp = 10)
    #################################################################################


    ################################################################################
              
    ################################################################################

    model_directory = '/home/kasiakm/Documents/QREM_DATA/Simulations100q/'
    

    ######################################################################################
    #### mitigation data creation                                       ##################
    ####specify number of circuits, shots and directory where data is to be saved#########
    #### number_of_benchmark_circuits : number of circuits used to perform mitigation  ###
    ######################################################################################
    data_directory = '/home/kasiakm/Documents/QREM_DATA/Simulations100q/'

    number_of_circuits = 1500

    number_of_shots = 10**4

    number_of_benchmark_circuits = 300

    for i in range(1):
        j=i+11
        model_file = '100qmodel'+str(j)+'.pkl'

        with open(model_directory+ model_file, 'rb') as filein:
            model_noise_dictionary = pickle.load(filein)
    
        model_noise_matrices_dictionary = model_noise_dictionary['noise_matrices']
        print('model dictionary loaded:',model_noise_matrices_dictionary.keys())
        print('dict:', model_noise_matrices_dictionary.keys())

        noise_model_simulation = cn_noise_model.CNModelData(number_of_qubits=number_of_qubits)
        noise_model_simulation.set_noise_model(model_noise_matrices_dictionary)
        realization = 'model'+str(j)+'_test'
        



    #characterization procedure is run, noise model data are computed
        noisy_experiment_dictionary = mitigation_routines.simulate_noisy_experiment(noise_model=noise_model_simulation,number_of_circuits=number_of_circuits,number_of_shots=number_of_shots,number_of_benchmark_circuits=number_of_benchmark_circuits,data_directory=data_directory,name_id = realization+str(number_of_circuits)+'x'+str(number_of_shots), return_ideal_experiment_data=False,save_data=True)

        noisy_results_dictionary = noisy_experiment_dictionary['noisy_results_dictionary']



from typing import Optional
from qrem.functions_qrem.ancillary_functions import get_local_storage_directory


#MOcomm - we do not seam to use these functions - hence I think the whole file can be deleted

def get_standarized_directory_main_tutorials(directory_QREM):
    return f"{directory_QREM}/Tutorials/data_storage"


def get_standarized_subdirectory_results_tutorials(experiment_name,
                                                    backend_name,
                                                    number_of_qubits,
                                                    shots_per_setting,
                                                    date_name):


    return f"{backend_name}/{experiment_name}/{date_name}/number_of_qubits_{number_of_qubits}/shots_{shots_per_setting}/"


def get_standarized_subdirectory_dot_collections_tutorials(experiment_name,
                                                    number_of_circuits,
                                                    number_of_qubits):


    return f"{experiment_name}/number_of_qubits_{number_of_qubits}/circuits_amount_{number_of_circuits}/"


def get_standarized_subdirectory_hamiltonians_tutorials(hamiltonian_name,
                                                    number_of_qubits):


    return f"{hamiltonian_name}/number_of_qubits_{number_of_qubits}/"




def _get_standard_subfolder_name(backend_name:str,
                              date:str,
                              number_of_qubits:int):


    return f"{backend_name}/{date}/number_of_qubits_{number_of_qubits}/"

def get_directory_circuits_collections(
        locality:int,
        number_of_qubits:int,
        circuits_amount:int,
        experiment_name='DDOT'):

    directory_to_go = get_local_storage_directory() + f"saved_data/circuits_collections/"
    directory_to_go = directory_to_go\
                      +f"{experiment_name}/locality_{locality}/number_of_qubits_{number_of_qubits}/circuits_amount_{circuits_amount}/"
    return directory_to_go





def get_directory_stored_hamiltonians(number_of_qubits:int,
                                      hamiltonian_name:str):

    directory_to_go = get_local_storage_directory() + f"saved_data/stored_hamiltonians/"
    directory_to_go = directory_to_go\
                      +f"{hamiltonian_name}/number_of_qubits_{number_of_qubits}/"
    return directory_to_go

def get_directory_raw_results(backend_name:str,
                              date:str,
                              number_of_qubits:int):

    directory_to_go = get_local_storage_directory() + f"saved_data/raw_experimental_results/"
    directory_to_go = directory_to_go+_get_standard_subfolder_name(backend_name=backend_name,
                                                                   date=date,
                                                                   number_of_qubits=number_of_qubits)

    return directory_to_go


def get_directory_backend_information(backend_name:str,
                              date:str):

    directory_to_go = get_local_storage_directory() + f"saved_data/backend_information/"
    directory_to_go = directory_to_go+f"{backend_name}/{date}/"
    return directory_to_go



def get_directory_processed_experimental_results(backend_name:str,
                                                 date:str,
                                                 number_of_qubits:int,
                                                 experiment_name:str,
                                                 additional_path:Optional[str]=None):


    directory_to_go = get_local_storage_directory()+f"saved_data/processed_experimental_results/{experiment_name}/"
    directory_to_go = directory_to_go+_get_standard_subfolder_name(backend_name=backend_name,
                                                                   date=date,
                                                                   number_of_qubits=number_of_qubits)

    if additional_path is None:
        additional_path = ''

    return directory_to_go+additional_path


"""
@authors: Filip Maciejewski, Oskar Słowik, Tomek Rybotycki
@contact: filip.b.maciejewski@gmail.com

REFERENCES:
[0] Filip B. Maciejewski, Zoltán Zimborás, Michał Oszmaniec,
"Mitigation of readout noise in near-term quantum devices
by classical post-processing based on detector tomography",
Quantum 4, 257 (2020)

[0.5] Filip B. Maciejewski, Flavio Baccari, Zoltán Zimborás, Michał Oszmaniec,
"Modeling and mitigation of cross-talk effects in readout noise
with applications to the Quantum Approximate Optimization Algorithm",
Quantum 5, 464 (2021).

"""
from qrem.functions_qrem import ancillary_functions as anf
import numpy as np
from typing import Optional, Dict, List, Union


class GlobalNoiseMatrixCreator:
    """
    This is class that, given noise matrices on clusters of qubits as function of input state of their
    neighbors, constructs global noise model on all qubits.
    """

    def __init__(self,
                 noise_matrices_dictionary: Dict[
                     str, Union[np.ndarray, Dict[str, Dict[str, np.ndarray]]]],
                 clusters_list: List[List[int]],
                 neighborhoods: Dict[str, List[int]]
                 ) -> None:

        self._noise_matrices_dictionary = noise_matrices_dictionary
        self._clusters_list = clusters_list

        for cluster_key, neighbors_list in neighborhoods.items():
            if neighbors_list is not None:
                if len(neighbors_list) == 0:
                    neighborhoods[cluster_key] = None

        self._neighborhoods = neighborhoods

        clusters_labels_list, neighbors_labels_list = [], []
        for cluster_index in range(len(self._clusters_list)):
            key_cluster = anf.get_qubits_keystring(self._clusters_list[cluster_index])
            clusters_labels_list.append(key_cluster)
            neighbors_labels_list.append(anf.get_qubits_keystring(self._neighborhoods[key_cluster]))

        self._clusters_labels_list = clusters_labels_list
        self._neighbors_labels_list = neighbors_labels_list

        self._matrix_elements_dictionary = {}
        self._number_of_qubits = sum([len(indices) for indices in clusters_list])
        self._global_noise_matrix = np.zeros((self._number_of_qubits, self._number_of_qubits),
                                             dtype=float)


    # #PPcomment: why not version from core/ancillary functions? MERGE
    # @staticmethod
    # def get_qubits_keystring(list_of_qubits):
    #     if list_of_qubits is None:
    #         return None
    #     return 'q' + 'q'.join([str(s) for s in list_of_qubits])

    def update_labels_lists(self):
        clusters_labels_list, neighbors_labels_list = [], []
        for cluster_index in range(len(self._clusters_list)):
            key_cluster = anf.get_qubits_keystring(self._clusters_list[cluster_index])
            clusters_labels_list.append(key_cluster)
            neighbors_labels_list.append(anf.get_qubits_keystring(self._neighborhoods[key_cluster]))

        self._clusters_labels_list = clusters_labels_list
        self._neighbors_labels_list = neighbors_labels_list

    def compute_matrix_element(self,
                               input_state: str,
                               output_state: str):
        """
        Function that computes single global noise matrix element.

       :param input_state: bitstring denoting INPUT classical state
       :param output_state: bitstring denoting OUTPUT classical state
        """

        matrix_element = 1
        for cluster_index in range(len(self._clusters_list)):
            cluster_label_now = self._clusters_labels_list[cluster_index]
            neighbors_now = self._neighborhoods[cluster_label_now]
            neighbors_label_now = self._neighbors_labels_list[cluster_index]
            qubits_now = self._clusters_list[cluster_index]



            if neighbors_now is None:
                neighbors_input_state_now = 'averaged'
            else:
                neighbors_input_state_now = ''.join([input_state[s] for s in neighbors_now])
            # print(neighbors_now,neighbors_input_state_now)

            cluster_input_state = ''.join([input_state[s] for s in qubits_now])
            cluster_output_state = ''.join([output_state[s] for s in qubits_now])

            if neighbors_label_now in self._noise_matrices_dictionary[cluster_label_now].keys():
                cluster_matrix = \
                    self._noise_matrices_dictionary[cluster_label_now][neighbors_label_now][
                        neighbors_input_state_now]
            else:
                if neighbors_now is None:
                    try:
                        cluster_matrix = self._noise_matrices_dictionary[cluster_label_now][
                            'averaged']
                    except(KeyError):
                        cluster_matrix = self._noise_matrices_dictionary[cluster_label_now][
                            '']
                else:
                    cluster_matrix = self._noise_matrices_dictionary[cluster_label_now][
                        neighbors_input_state_now]

            matrix_element *= cluster_matrix[int(cluster_output_state, 2),
                                             int(cluster_input_state, 2)]

        return matrix_element

    def compute_global_noise_matrix(self):
        """
            This method_name is faster than other one
        """
        number_of_qubits = self._number_of_qubits

        dimension = int(2 ** number_of_qubits)

        classical_register = anf.get_classical_register_bitstrings(range(number_of_qubits))

        self._global_noise_matrix = np.zeros((dimension, dimension))

        for input_state_bitstring in classical_register:
            for output_state_bitstring in classical_register:
                self._global_noise_matrix[int(output_state_bitstring, 2),
                                          int(input_state_bitstring, 2)] = \
                    self.compute_matrix_element(input_state_bitstring, output_state_bitstring)

        return self._global_noise_matrix

    def compute_global_noise_matrix_old(self,
                                        clusters_list: Optional[List[List[int]]] = None,
                                        neighbors_of_clusters: Optional[List[List[int]]] = None):
        updated_lists = False
        if clusters_list is None:
            if self._clusters_list is None:
                raise ValueError('Please provide clusters list.')
            else:
                self.update_labels_lists()
                updated_lists = True

            clusters_list = self._clusters_list

        if neighbors_of_clusters is None:
            if self._neighborhoods is None:
                raise ValueError('Please provide neighbors list')
            else:
                if not updated_lists:
                    self.update_labels_lists()

                neighbors_of_clusters = [self._neighborhoods[anf.get_qubits_keystring(clust)] for clust in
                                         clusters_list]

        lambdas = [self._noise_matrices_dictionary[anf.get_qubits_keystring(clust)] for clust in
                   clusters_list]
        number_of_qubits = sum([len(inds) for inds in clusters_list])

        d = int(2 ** number_of_qubits)

        big_lambda = np.zeros((d, d))

        for input_state_integer in range(d):
            input_state_bitstring = "{0:b}".format(input_state_integer).zfill(number_of_qubits)
            for output_state_integer in range(d):
                output_state_bitstring = "{0:b}".format(output_state_integer).zfill(number_of_qubits)

                element = 1
                for cluster_index in range(len(lambdas)):
                    qubits_now = clusters_list[cluster_index]
                    neighbours_now = neighbors_of_clusters[cluster_index]

                    if neighbours_now is not None:
                        input_state_neighbors = ''.join(
                            [input_state_bitstring[a] for a in neighbours_now])
                        neighbors_string = anf.get_qubits_keystring(neighbours_now)
                        if neighbors_string in lambdas[cluster_index].keys():
                            lambda_of_interest = lambdas[cluster_index][neighbors_string][
                                input_state_neighbors]
                        else:
                            lambda_of_interest = lambdas[cluster_index][
                                input_state_neighbors]
                    else:
                        try:
                            lambda_of_interest = lambdas[cluster_index]['averaged']
                        except(KeyError):
                            try:
                                lambda_of_interest = lambdas[cluster_index]['']
                            except(KeyError):
                                raise KeyError('Something wrong with averaged lambda')

                    small_string_ideal = ''.join(
                        [list(input_state_bitstring)[b] for b in qubits_now])
                    small_string_measured = ''.join(
                        [list(output_state_bitstring)[b] for b in qubits_now])

                    element *= lambda_of_interest[
                        int(small_string_measured, 2), int(small_string_ideal, 2)]

                big_lambda[output_state_integer, input_state_integer] = element

        return big_lambda



""" Needs to be refactored into somewhere else, """

from typing import List, Dict, Union, Optional
import numpy as np
from tqdm import tqdm
from qiskit.result import Result

from qrem.common.printer import qprint
from qrem.functions_qrem.povmtools import reorder_probabilities

def get_shots_results_from_qiskit_result(result_object):
    pass


# ORGANIZE: Three functions below are copied from qiskit_utilities.
# TODO: check if job_headers are needed anywhere

def get_frequencies_array_from_probabilities_list(frequencies_list:List[Union[List[float], np.ndarray]],
                                                  reverse_order:Optional[bool]=False):
    """
    :param frequencies_list: list of probabilities estimated in experiments
    :param reverse_order: specify whether probabiltiies lists should be reordered, which corresponds
                          to changing qubits' ordering

    """

    number_of_qubits = int(np.log2(len(frequencies_list[0])))
    frequencies_formatted = frequencies_list
    if reverse_order:
        frequencies_formatted = [reorder_probabilities(probs, range(number_of_qubits)[::-1])
                                  for probs in frequencies_list]

    frequencies_array = np.ndarray(shape=(len(frequencies_list), len(frequencies_list[0])))

    for probe_state_index in range(len(frequencies_list)):
        frequencies_array[probe_state_index][:] = frequencies_formatted[:]
    return frequencies_array


def get_frequencies_array_from_results(results_list: List[Result]) -> np.ndarray:
    """
    Description:
        Creates an array of frequencies from given qiskit job results. This method is working with
        qiskit 0.16. The shape of the array is
            c x 2 ** q,
        where c denotes circuits number and q denotes number of qubits.
    Parameters:
        :param results_list: List of qiskit jobs_list results.
    Returns:
        np.ndarray with shape=0 if there were no circuits in the job, or with shape c x 2 ** q
        containing frequencies data for each possible state.
    Notes:
        Possible states are numbered increasingly from |00000 ... 0>, |10000 ... 0> up to |1111 ... 1>.
    """

    all_circuits_number = sum(len(results.results) for results in results_list)

    if all_circuits_number == 0:
        return np.ndarray(shape=0)

    # The length of a state describes how many qubits were used during experiment.
    number_of_qubits = len(next(iter(results_list[0].get_counts(0).keys())))

    classical_register = ["{0:b}".format(i).zfill(number_of_qubits) for i in range(2 ** number_of_qubits)]
    frequencies_array = np.ndarray(shape=(all_circuits_number, len(classical_register)))

    # TODO TR: This has to be rewritten as it's too nested.
    for results in results_list:
        number_of_circuits_in_results = len(results.results)
        for i in range(number_of_circuits_in_results):
            counts = results.get_counts(i)
            shots_number = results.results[i].shots

            # TODO FBM: added here accounting for reversed register in qiskit
            normal_order = []
            for j in range(len(classical_register)):
                if classical_register[j] in counts.keys():
                    normal_order.append(counts[classical_register[j]] / shots_number)
                else:
                    normal_order.append(0)
            # frequencies = reorder_probabilities(normal_order, range(states_len))
            frequencies = reorder_probabilities(normal_order, range(number_of_qubits)[::-1])

            frequencies_array[i][:] = frequencies[:]

    return frequencies_array


def get_counts_from_result_object(results_object,
                                  counts_dictionary_to_update={},
                                  circuit_names_list=None,
                                  job_index=0):
    """
    This functions takes a qiskit Result object and returns a dictionary with counts for each Experiment in this object.
    returns:
    counts_dictionary_to_update[str, counts]: the key is the name of a circuit, the value is a dictionary of counts,
    where the key is a string denoting classical outcome and the value the number of its occurrences in given experiment.
    The circuits are named according to convention from get_circuits in SeparableCircuitsCreator:
    circuit_name = "experiment name" + "-" + "circuit label" +
        "no" + "integer identifier for multiple implementations of the same circuit", e.g.: "DDOT-010no3"
    """
    number_of_experiments = len(list(results_object.results))
    for exp_index in tqdm(range(number_of_experiments)):
        try:
            circuit_name_now = results_object.results[exp_index].header.name
            counts_now = results_object.get_counts(circuit_name_now)
        except (KeyError, AttributeError) as warn:
            # print(warn)
            if circuit_names_list is None:
                qprint(f"Experiment: no {exp_index} does not have a header. "
                               f"using_default_name")
                circuit_name_now = f"default_name_no_{job_index}-{exp_index}"
                counts_now = results_object.get_counts(exp_index)
            else:
                circuit_name_now = circuit_names_list[exp_index]
                counts_now = results_object.get_counts(exp_index)
        counts_dictionary_to_update[circuit_name_now] = counts_now
    return counts_dictionary_to_update

#TODO: check if job_headers are needed anywhere
def get_counts_from_qiskit_jobs(jobs_list,
                         return_job_headers=False) -> Dict[str, Dict[str, int]]:
    """
    This function takes a list of qiskit jobs (e.g. what 'run_batches' returns) and returns a dictionary of outcomes of
    each of the circuits in the list, where keys are circuit names and values are dictionaries of counts (where the key
    is a string denoting classical outcome and the value the number of its occurrences in given experiment)

    """
    qprint('Getting counts...')
    counts_dictionary, job_headers = {}, []

    for job_index in tqdm(range(len(jobs_list))):
        job_now = jobs_list[job_index]
        job_header_now = job_now.result().qobj_id
        job_headers.append(job_header_now)
        results_object_now = job_now.result()
        circuit_name_list = None
        if hasattr(job_now, 'metadata'):
            if 'circuit_names' in job_now.metadata:
                circuit_name_list = job_now.metadata['circuit_names']  # this attribute exists only for braket jobs

        counts_dictionary = get_counts_from_result_object(results_object=results_object_now,
                                                          counts_dictionary_to_update=counts_dictionary,
                                                          circuit_names_list=circuit_name_list,
                                                          job_index=job_index)
    qprint("GOT IT!")

    if return_job_headers:
        return counts_dictionary, job_headers
    else:
        return counts_dictionary



import numpy as np
from scipy.stats import norm
import random


""" Generates block correlation table with constant values within each block and constant bias value outside
    all blocks (clusters).
"""
def generate_simple_correlations_table(block_sizes, values, bias):
    total_size = 0
    for size in block_sizes:
        total_size+=size

    correlations_table=np.full((total_size, total_size), bias)

    shift=0
    for block_no in range(len(block_sizes)):
        for local_index_i in range(block_sizes[block_no]):
            for local_index_j in range(block_sizes[block_no]):
                if local_index_i != local_index_j:
                    correlations_table[local_index_i+shift][local_index_j+shift]=values[block_no]
                else:
                    correlations_table[local_index_i+shift][local_index_j+shift]=0
        shift+= block_sizes[block_no]

    return correlations_table

""" Generates block correlation table with off-diagonal values within each block drawn from gaussian distribution centered
 halfway between treshold and 1. Values outside all block are drawn from gaussian distribution centered
 halfway between 0 and treshold. Values outside the range [0,1] are rejected and chosen until they are in [0,1].
Overlap parameter regulates the (common) variance of both gaussians so that the max from probabilities of getting off-block value
 larger than treshold and in-block value below the treshold is = overlap.
"""
def generate_simple_correlations_table_gaussian(block_sizes, treshold, overlap):
    mu_block = 0.5 + treshold/2.0
    mu_rest = treshold/2.0

    sigma=0
    if treshold < 0.5:
        sigma = (treshold-mu_rest)/norm.ppf(1-overlap)
    else:
        sigma = (treshold-mu_block)/norm.ppf(overlap)

    total_size = 0
    for size in block_sizes:
        total_size += size

    correlations_table = np.full((total_size, total_size), 0.0)

    shift = 0
    for block_no in range(len(block_sizes)):
        for local_index_i in range(block_sizes[block_no]):
            for local_index_j in range(block_sizes[block_no]):
                if local_index_i != local_index_j:
                    while True:
                        val=random.gauss(mu_block, sigma)
                        correlations_table[local_index_i + shift][local_index_j + shift] = val
                        if val >= 0.0 and val <=1.0:
                            break
                else:
                    correlations_table[local_index_i + shift][local_index_j + shift] = 0


        for local_index_i in range(block_sizes[block_no]):
            for local_index_j in range(block_sizes[block_no],total_size-shift):
                while True:
                    val = random.gauss(mu_rest, sigma)
                    correlations_table[local_index_i + shift][local_index_j + shift] = val
                    if val >= 0.0 and val <=1.0:
                        break

        for local_index_i in range(block_sizes[block_no],total_size-shift):
            for local_index_j in range(block_sizes[block_no]):
                while True:
                    val = random.gauss(mu_rest, sigma)
                    correlations_table[local_index_i + shift][local_index_j + shift] = val
                    if val >= 0.0 and val <=1.0:
                        break
        shift += block_sizes[block_no]
    return correlations_table


"""


Adapted from:
https://github.com/Hannoskaj/Hyperplane_Intersection_Projection

Based on reference:
Projected Least-Squares Quantum Process Tomography
Trystan Surawy-Stepney, Jonas Kahn, Richard Kueng, and Madalin Guta,
https://arxiv.org/pdf/2107.01060.pdf


"""

import numpy as N
import scipy.linalg as SL



def proj_TP(rho):
    """
    Projects the Choi matrix rho of a channel on trace-preserving channels.
    """
    d = N.sqrt(len(rho)).astype(int)
    partial_mixed = N.eye(d) / d

    # N.trace on the axes corresponding to the system
    correction = N.einsum('de, fg -> dfeg',partial_mixed, (partial_mixed - N.trace(rho.reshape(4 * [d]), axis1=0, axis2=2)))
    return rho + correction.reshape(d**2, d**2)




def final_CPTP_by_mixing(rho, full_output=False):
    """
    Assumed to be in TP.
    """
    d = len(rho)
    abs_least_ev = - SL.eigvalsh(rho, subset_by_index=[0,0])[0]
    if full_output:
        return (rho + abs_least_ev * N.eye(d)) / (1 + d * abs_least_ev), - abs_least_ev
    else:
        return (rho + abs_least_ev * N.eye(d)) / (1 + d * abs_least_ev)





def ensure_trace(eigvals):
    """
    Assumes sum of eigvals is at least one.

    Finds the value l so that $\sum (\lambda_i - l)_+ = 1$
    and set the eigenvalues $\lambda_i$ to $(\lambda_i - l)_+$.
    """
    trace = eigvals.sum()
    while trace > 1:
        indices_positifs = eigvals.nonzero()
        l = len(indices_positifs[0]) # Number of (still) nonzero eigenvalues
        eigvals[indices_positifs] += (1 - trace) / l
        eigvals = eigvals.clip(0)
        trace = eigvals.sum()
    return eigvals



def new_proj_CP_threshold(rho,  free_trace=True, full_output=False, thres_least_ev=False):
    """
    If thres_least_ev=False and free_trace=False, then projects rho on CP
    trace_one operators.

    More generally, changes the eigenvalues without changing the eigenvectors:
    * if free_trace=True and thres_least_ev=False, then projects on CP operators,
    with no trace condition.
    * if thres_least_ev=True, free_trace is ignored. Then we threshold by minus
    the least eigenvalues before projecting on CP trace-one operator, if we
    can do that without modifying any eigenvalue by more than threshold. If we
    cannot, we increase the largest eigenvalues by threshold, until we arrive at
    trace one. The eigenvalue that allows passing 1 is set to the value to get a
    sum of exactly one, and all the remaining ones are set to zero.
    """
    eigvals, eigvecs = SL.eigh(rho) # Assumes hermitian; sorted from lambda_min to lambda_max

    least_ev = eigvals[0]
    
    from qrem.functions_qrem import ancillary_functions as anf
    
    # qprint_array(rho,5)

    if thres_least_ev:
        threshold = - least_ev # > 0
        high_indices = N.where(eigvals > threshold)
        low_indices = N.where(eigvals <= threshold)
        if (eigvals[high_indices] + threshold).sum() > 1:
            eigvals[low_indices] = 0
            eigvals[high_indices] += threshold
            eigvals = ensure_trace(eigvals)
        else:
            eigvals += threshold
            inv_cum_evs = eigvals[::-1].cumsum()[::-1]
            first_less_1 = N.where(inv_cum_evs < 1)[0][0]
            eigvals[:first_less_1 - 1] = 0
            eigvals[first_less_1 - 1] = 1 - inv_cum_evs[first_less_1]

    else:
        eigvals = eigvals.clip(0)
        if not free_trace:
            eigvals = ensure_trace(eigvals)
        #
    indices_positifs = eigvals.nonzero()[0]
    rho_hat_TLS = (eigvecs[:,indices_positifs] * eigvals[indices_positifs]) @ eigvecs[:,indices_positifs].T.conj()
    
    # import numpy as np
    # print('this')
    # print(rho_hat_TLS)
    # qprint_array(np.array(rho_hat_TLS),5)
    # print("""
    #
    # """)

    if full_output==2:
        return rho_hat_TLS, least_ev, len(indices_positifs)
    elif full_output:
        return rho_hat_TLS, least_ev
    else:
        return rho_hat_TLS




def proj_CP_threshold(rho,  free_trace=True, full_output=False, thres_least_ev=False):
    """
    If thres_least_ev=False and free_trace=False, then projects rho on CP
    trace_one operators.

    More generally, changes the eigenvalues without changing the eigenvectors:
    * if free_trace=True and thres_least_ev=False, then projects on CP operators,
    with no trace condition.
    * if thres_least_ev=True, free_trace is ignored. Then we bound from below all
    eigenvalues by their original value plus the least eigenvalue (which is negative).
    Then all the lower eigenvalues take the lower bound (or zero if it is negative),
    all the higher eigenvalues are unchanged, and there is one eigenvalue in the middle
    that gets a value between its lower bound and its original value, to ensure the
    trace is one.
    """
    eigvals, eigvecs = SL.eigh(rho) # Assumes hermitian; sorted from lambda_min to lambda_max

    least_ev = eigvals[0]

    if thres_least_ev:
        threshold = - least_ev # > 0
        evlow = (eigvals - threshold).clip(0)
        toadd = eigvals - evlow
        missing = 1 - evlow.sum()
        if missing < 0: # On this rare event, revert to usual projection
            eigvals = eigvals.clip(0)
            eigvals = ensure_trace(eigvals)
        else:
            inv_cum_toadd =  toadd[::-1].cumsum()[::-1]
            last_more_missing = N.where(inv_cum_toadd >= missing)[0][-1]
            eigvals[:last_more_missing] = evlow[:last_more_missing]
            eigvals[last_more_missing] = eigvals[last_more_missing] + missing - inv_cum_toadd[last_more_missing]
    else:
        eigvals = eigvals.clip(0)
        if not free_trace:
            eigvals = ensure_trace(eigvals)
        #
    indices_positifs = eigvals.nonzero()[0]
    rho_hat_TLS = (eigvecs[:,indices_positifs] * eigvals[indices_positifs]) @ eigvecs[:,indices_positifs].T.conj()

    if full_output==2:
        return rho_hat_TLS, least_ev, len(indices_positifs)
    elif full_output:
        return rho_hat_TLS, least_ev
    else:
        return rho_hat_TLS





def step2(XW, target):
    """
    Finds a (big) subset of hyperplanes, including the last one, such that
    the projection of the current point on the intersection of the corresponding
    half-spaces is the projection on the intersection of hyperplanes.

    Input: XW is the matrix of the scalar products between the different
    non-normalized normal directions projected on the subspace TP, written w_i
    in the main functions.
    target is the intercept of the hyperplanes with respect to the starting point,
    on the scale given by w_i.

    Outputs which hyperplanes are kept in subset, and the coefficients on their
    respective w_i in coeffs.
    """
    nb_active = XW.shape[0]
    subset = N.array([nb_active - 1])
    coeffs = [target[-1] / XW[-1, -1]] # Always positive
    for i in range(nb_active - 2, -1, -1):
        test = (XW[i, subset].dot(coeffs) < target[i])
        # The condition to project on the intersection of the hyperplanes is that
        # all the coefficients are non-negative. This is equivalent to belonging
        # to the normal cone to the facet.
        if test:
            subset = N.r_[i, subset]
            coeffs = SL.inv(XW[N.ix_(subset, subset)]).dot(target[subset])
            # Adding a new hyperplane might generate negative coefficients.
            # We remove the corresponding hyperplanes, except if it is the last
            # hyperplane, in which case we do not add the hyperplane.
            if coeffs[-1] < 0:
                subset = subset[1:]
                coeffs = SL.inv(XW[N.ix_(subset, subset)]).dot(target[subset])
            elif not N.all(coeffs >= 0):
                subset = subset[N.where(coeffs >= 0)]
                coeffs = SL.inv(XW[N.ix_(subset, subset)]).dot(target[subset])

    return subset, coeffs




def la(XWb, sq_norm_xn):
    target = N.zeros((XWb.shape[0],))
    target[-1] = sq_norm_xn
    return SL.inv(XWb).dot(target)

def step(XW, sq_norm_xn):
    nb_active = XW.shape[0]
    subset = [nb_active - 1]
    coeffs = [sq_norm_xn / XW[-1, -1]] # Always positive
    for i in range(nb_active - 2, -1, -1):
        test = (XW[i, subset].dot(coeffs) < 0)
        if test:
            subset = [i] + subset
            coeffs = la(XW[N.ix_(subset, subset)], sq_norm_xn) # Always positive ??? Vérifier
           # assert N.all(coeffs >= 0)
           # for now, print and correct and continue
            if not N.all(coeffs >= 0):
                print('There seems to be a negative coefficient')
                print(coeffs)
                print('The offending w is removed.')
                subset = subset[1:]
                coeffs = la(XW[N.ix_(subset, subset)], sq_norm_xn)
    return subset, coeffs







def HIP_switch(rho, HIP_to_alt_switch='first', alt_to_HIP_switch='cos', maxiter=200, depo_tol=1e-3,
        depo_rtol=1e-1, min_cos = .99, alt_steps=4, missing_w=1, min_part=.3, HIP_steps=10):

    dim2 = len(rho)
    sel = 'alternate' # Selector for the step; 'alternate' or 'HIP'.
    if alt_to_HIP_switch == 'cos':
        w_norm_ancien = N.zeros((dim2, dim2)) # Not normalized to ensure at least two steps are taken.
    elif alt_to_HIP_switch == 'counter':
        past_al = 0       # number of steps already made in 'alternate' mode.
        alt_step_gen = step_generator(alt_steps)
        current_alt_step = next(alt_step_gen)
    else:
        raise ValueError('Unknown alt_to_HIP_switch. Must be "cos" or "counter".')

    if HIP_to_alt_switch == 'counter':
        HIP_step_gen = step_generator(HIP_steps)
        past_HIP = 0
    elif HIP_to_alt_switch == 'part':
        pass
    elif HIP_to_alt_switch == 'first':
        pass
    elif HIP_to_alt_switch == 'missing':
        missed = 0
    else:
        raise ValueError('Unknown HIP_to_alt_switch. Must be "first", "missing", "part" or "counter".')



    dims = (dim2, dim2)

    active = N.array([])
    nb_actives = 0
    XW = N.zeros((0,0))
    w_act = N.zeros([0, dim2, dim2])
    target = N.array([])
    coeffs = N.array([])

    # rho is on CP, we first project on TP. Outside the loop because we also end on TP.
    rho = proj_TP(rho)

    for m in range(maxiter):

        # On CP
        rho_after_CP, least_ev = proj_CP_threshold(rho, free_trace, full_output=True)

        # Breaks here because the (- least_ev) might increase on the next rho
        if  (- least_ev) < least_ev_x_dim2_tol / dim2:
            break

        if sel == 'alternate':

            # On TP and intersection with hyperplane
            if alt_to_HIP_switch == 'cos':
                w_new = proj_TP(rho_after_CP) - rho
                norm_w = SL.norm(w_new)
                change = (N.vdot(w_new / norm_w, w_norm_ancien).real > min_cos)
                w_norm_ancien = w_new / norm_w

                # If change with alt_steps, the current projection is transformed into
                # the first HIP step.
                if change:
                    active = N.array([m])
                    nb_actives = 1
                    XW = N.array([[norm_w**2]])
                    w_act = N.array([w_new])
                    coeffs = N.array([SL.norm(rho - rho_after_CP)**2 / norm_w**2])
                    target = N.array([0.])
                    rho += coeffs[0] * w_new

                else:
                    rho += w_new

            elif alt_to_HIP_switch == 'counter':
                rho = proj_TP(rho_after_CP)
                past_al += 1
                change = (past_al >= current_alt_step)

                if change:
                    active = N.array([])
                    nb_actives = 0
                    XW = N.zeros((0,0))
                    w_act = N.zeros([0, dim2, dim2])
                    target = N.array([])
                    coeffs = N.array([])

            if change:
                if HIP_to_alt_switch == 'missing':
                    missed = 0
                elif HIP_to_alt_switch == 'counter':
                    past_HIP = 0
                    current_HIP_step = next(HIP_step_gen)
                sel = 'HIP'


        elif sel == 'HIP': # No other possibility

            sq_norm_x_i = SL.norm(rho - rho_after_CP)**2
            w_i =  proj_TP(rho_after_CP) - rho
            xiwi = SL.norm(w_i)**2

            XW = N.column_stack([XW, N.zeros(nb_actives)])
            XW = N.row_stack([XW, N.zeros(nb_actives + 1)])
            new_xw = N.einsum('ij, kij -> k', w_i.conj(), w_act).real # Notice that the scalar product are all real
                                                                      # since the matrices are self-adjoint.
            XW[-1, :-1] = new_xw
            XW[:-1, -1] = new_xw
            XW[-1, -1]  = xiwi
            target = N.r_[target, sq_norm_x_i]


            active = N.concatenate((active, [m]))
            w_act = N.concatenate([w_act, [w_i]])

            subset, coeffs = step2(XW, target)

            if HIP_to_alt_switch == 'missing':
                missed += len(active) - len(subset) # Don't move this after the update to active !!!

            XW = XW[N.ix_(subset, subset)]
            active = active[subset]
            nb_actives = len(active)
            w_act = w_act[subset]
            target = N.zeros((nb_actives,))
            rho += N.einsum('k, kij -> ij', coeffs, w_act)

            if HIP_to_alt_switch in ['first', 'part']:
                if (subset[0] != 0):
                    change = True
                elif HIP_to_alt_switch == 'part':
                    step_size = N.sqrt(N.einsum('i, ij, j', coeffs, XW, coeffs))
                    w_first_contrib = coeffs[0] * N.sqrt(XW[0,0])
                    change = (min_part * step_size >= w_first_contrib)
                else:
                    change = False
            elif HIP_to_alt_switch == 'missing':
                change = (missed >= missing_w)
            elif HIP_to_alt_switch == 'counter':
                past_HIP += 1
                change = (past_HIP >= current_HIP_step)

            if change:
                if alt_to_HIP_switch == 'cos':
                    w_norm_ancien = N.zeros((dim2, dim2)) # Ensures two alternate steps. Also possible to
                                                          # use w_norm_ancien = w_i / N.sqrt(xiwi)
                elif alt_to_HIP_switch == 'counter':
                    past_al = 0
                    current_alt_step = next(alt_step_gen)
                sel = 'alternate'

    return rho




def increasing_steps(step, start=None, maxi=N.inf):
    """
    Yields a generator, increasing each output by the same.
    If start is given, starts at start instead of step.
    If maxi is given, always yields maxi when the sum of steps exceeds maxi.
    """
    if start is None:
        res = step
    else:
        res = start
    while res < maxi:
        yield res
        res += step
    while True:
        yield maxi





"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Tomasz Rybotycki, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com

References:
[0] Filip B. Maciejewski, Zoltán Zimborás, Michał Oszmaniec,
"Mitigation of readout noise in near-term quantum devices
by classical post-processing based on detector tomography",
Quantum 4, 257 (2020)

[0.5] Filip B. Maciejewski, Flavio Baccari, Zoltán Zimborás, Michał Oszmaniec,
"Modeling and mitigation of cross-talk effects in readout noise
with applications to the Quantum Approximate Optimization Algorithm",
Quantum 5, 464 (2021).

[2] Zbigniew Puchała, Łukasz Pawela, Aleksandra Krawiec, Ryszard Kukulski, "Strategies for optimal single-shot
discrimination of quantum measurements", Phys. Rev. A 98, 042103 (2018), https://arxiv.org/abs/1804.05856

[3] T. Weissman, E. Ordentlich, G. Seroussi, S. Verdul, and M. J. Weinberger, Technical Report HPL-2003-97R1,
Hewlett-Packard Labs (2003).

[4] John Smolin, Jay M. Gambetta, Graeme Smith, "Maximum Likelihood, Minimum Effort", Phys. Rev. Lett. 108, 070502
(2012), https://arxiv.org/abs/1106.5458
"""
import copy
from typing import List, Optional, Iterable

import cmath as c
import numpy as np
import numpy.typing as npt
import scipy as sc

import math
from enum import Enum
from scipy import optimize as scopt
import picos

from qrem.functions_qrem import ancillary_functions as anf
from qrem.functions_qrem import functions_distances as fun_dist


from qrem.common import utils, convert, math as qmath
# TODO_PP >> Talk with Paweł about import shortcut convention
from qrem.common.math import Constants as Const
from qrem.common.constants import SIGNIFICANT_DIGITS_POVM
from qrem.common.math import identity_check

try:
    import qiskit
except(ModuleNotFoundError):
    pass


threshold = 10 ** (-SIGNIFICANT_DIGITS_POVM)


def check_if_projector_is_in_computational_basis(projector: npt.ArrayLike, d=2):
    n = int(math.log(np.shape(projector)[0], d))
    computational_basis_projectors = computational_projectors(d, n)

    for base_projector in computational_basis_projectors:
        projectors_difference = base_projector - projector  # independent from global phase
        norm = np.linalg.norm(projectors_difference)
        if abs(norm) < threshold:
            return True

    return False


def computational_projectors(d, n=1):
    return [get_density_matrix(computational_basis(d, n)[i]) for i in range(d ** n)]


def computational_basis(d, n=1):
    m_d = d ** n
    eye = np.eye(m_d)
    return [np.array(eye[:, i]).reshape(m_d, 1) for i in range(m_d)]


def get_density_matrix(ket):
    bra = np.matrix.getH(ket)
    return ket @ bra


# TODO TR: Check whether variables has been properly renamed.
def spectral_decomposition(matrix: np.ndarray, hermitian=False):
    if hermitian:
        eigenvalues, eigenvalues_matrix = np.linalg.eigh(matrix)
    else:
        eigenvalues, eigenvalues_matrix = np.linalg.eig(matrix)

    dimension = matrix.shape[0]
    projectors = [qmath.outer_product(
        np.array(eigenvalues_matrix[:, i]).reshape(dimension, 1)) for i in
        range(dimension)]

    return eigenvalues, projectors


# def get_unitary_change_state(state):
#     if check_if_projector_is_in_computational_basis(state):
#         if state[0][0] == 1:
#             return np.eye(2)
#         elif state[1][1] == 1:
#             return anf.pauli_sigmas['X']
#         else:
#             raise ValueError('error')
#     else:
#         m_a, m_u = spectral_decomposition(state)

#         d = m_u.shape[0]
#         determinant = np.linalg.det(m_a)

#         delta = c.phase(determinant) / d

#         m_u = c.exp(-1j * delta) * m_u

#         return np.matrix.getH(m_u)


def euler_angles_1q(unitary_matrix):
    # TODO FBM: This is slightly modified copied qiskit function
    _CUTOFF_PRECISION = 10 ** (-7)

    """Compute Euler angles for single-qubit gate.

    Find angles (theta, phi, lambda) such that
    unitary_matrix = phase * Rz(phi) * Ry(theta) * Rz(lambda)

    Args:
        unitary_matrix (ndarray): 2x2 unitary matrix

    Returns:
        tuple: (theta, phi, lambda) Euler angles of SU(2)

    Raises:
        QiskitError: if unitary_matrix not 2x2, or failure
    """
    if unitary_matrix.shape != (2, 2):
        raise ValueError("euler_angles_1q: expected 2x2 matrix")

    import scipy.linalg as la
    phase = la.det(unitary_matrix) ** (-1.0 / 2.0)
    U = phase * unitary_matrix  # U in SU(2)
    # OpenQASM SU(2) parameterization:
    # U[0, 0] = exp(-i_index(phi+lambda)/2) * cos(theta/2)
    # U[0, 1] = -exp(-i_index(phi-lambda)/2) * sin(theta/2)
    # U[1, 0] = exp(i_index(phi-lambda)/2) * sin(theta/2)
    # U[1, 1] = exp(i_index(phi+lambda)/2) * cos(theta/2)
    theta = 2 * math.atan2(abs(U[1, 0]), abs(U[0, 0]))

    # Find phi and lambda
    phiplambda = 2 * np.angle(U[1, 1])
    phimlambda = 2 * np.angle(U[1, 0])
    phi = (phiplambda + phimlambda) / 2.0
    lamb = (phiplambda - phimlambda) / 2.0

    # Check the solution
    Rzphi = np.array([[np.exp(-1j * phi / 2.0), 0],
                      [0, np.exp(1j * phi / 2.0)]], dtype=complex)
    Rytheta = np.array([[np.cos(theta / 2.0), -np.sin(theta / 2.0)],
                        [np.sin(theta / 2.0), np.cos(theta / 2.0)]], dtype=complex)
    Rzlambda = np.array([[np.exp(-1j * lamb / 2.0), 0],
                         [0, np.exp(1j * lamb / 2.0)]], dtype=complex)
    V = np.dot(Rzphi, np.dot(Rytheta, Rzlambda))
    if la.norm(V - U) > _CUTOFF_PRECISION:
        raise ValueError("compiling.euler_angles_1q incorrect result norm(V-U)={}".
                         format(la.norm(V - U)))
    return theta, phi, lamb


# MOVE_TO >> core.quantum
def get_su2_parametrizing_angles(m_a):
    # Get three angles theta, phi, lambda which parametrize single-qubit unitary from SU(2)

    if qmath.is_pauli_x(m_a):
        return ['x']
    elif qmath.identity_check(m_a):
        return ['id']

    m_a = qmath.round_array_to_ndigits(m_a)

    eps_angles = 10 ** (-7)
    determinant = np.linalg.det(m_a)
    delta = c.phase(determinant) / 2
    m_a = c.exp(-1j * delta) * m_a

    decomposer = qiskit.quantum_info.synthesis.one_qubit_decompose.OneQubitEulerDecomposer()

    euler_theta_phi_lambda = decomposer.angles(m_a)

    angles = [euler_theta_phi_lambda[0], euler_theta_phi_lambda[1], euler_theta_phi_lambda[2]]

    for i in range(0, 3):
        if abs(angles[i]) < eps_angles:
            angles[i] = 0

    return angles


# prepare unitaries which are to be implemented to change computational basis initial state of qubits (assuming
# |0000...0> initial state) for chosen probe states. Here we assume perfect qubit initialization.
def get_unitary_change_ket_qubit(ket):
    state = get_density_matrix(ket)
    if check_if_projector_is_in_computational_basis(state):
        if state[0][0] == 1:
            return np.eye(2)
        elif state[1][1] == 1:
            return Const.pauli_sigmas()["X"]
        else:
            raise ValueError('error')
    else:
        U = np.zeros((2, 2), dtype=complex)
        U[:, 0] = ket[:, 0]
        ket_comp = np.array([[1], [0]]).reshape(2, 1)
        ket_perp = ket_comp - np.vdot(ket_comp, ket) * ket
        ket_perp = ket_perp / np.linalg.norm(ket_perp)
        U[:, 1] = ket_perp[:, 0]

        determinant = np.linalg.det(U)
        delta = c.phase(determinant) / 2

        U = c.exp(-1j * delta) * U

        return U


# PP wht not use enumerated_dict function from core /ancillaryfunctions?
def get_enumerated_rev_map_from_indices(indices):
    # TODO: move this function somewhere else
    enumerated_dict = dict(enumerate(indices))
    rev_map = {}
    for k, v in enumerated_dict.items():
        rev_map[v] = k
    return rev_map


# MOVE_TO >> core.utils core.numerics // DUPLICATE - remove less relevant one
def all_possible_bitstrings_of_length(n: int, rev=True, form=str):
    """Generate outcome bitstrings for number_of_qubits-qubits.

    Args:
        n (int): the number of qubits.

    Returns:
        list:  list of bitstrings ordered as follows:
        Example: number_of_qubits=2 returns ['00', '01', '10', '11'].
"""
    if (form == str):
        if (rev == True):
            return [(bin(j)[2:].zfill(n))[::-1] for j in list(range(2 ** n))]
        else:
            return [(bin(j)[2:].zfill(n)) for j in list(range(2 ** n))]
    elif (form == list):
        if (rev == True):

            return [(list(bin(j)[2:].zfill(n))[::-1]) for j in list(range(2 ** n))]
        else:
            return [(list(bin(j)[2:].zfill(n))) for j in list(range(2 ** n))]


# MOVE_TO >> core.utils core.numerics // DUPLICATE
def register_names_qubits(qs,
                          qrs,
                          rev=False):
    # TODO: move this function somewhere else
    if qrs == 0:
        return ['']

    if (qrs == 1):
        return ['0', '1']

    all_names = all_possible_bitstrings_of_length(qrs, rev)
    not_used = []

    for j in list(range(qrs)):
        if j not in qs:
            not_used.append(j)

    bad_names = []
    for name in all_names:
        for k in (not_used):
            rev_name = name[::-1]
            if (rev_name[k] == '1'):
                bad_names.append(name)

    relevant_names = []
    for name in all_names:
        if name not in bad_names:
            relevant_names.append(name)

    return relevant_names


def calculate_total_variation_distance(p: np.array, q: np.array) -> float:
    """
    Description:
        Given two vectors calculate Total-Variation distance between them. See Refs. [1] and [2] for the relation
        between TV-distance and operational distance between quantum measurements.

    Parameters:
        :param p: numpy vector
        :param q: numpy vector

    Returns:
         Total variation distance between vectors q and p.
    """

    return np.linalg.norm(p - q, ord=1) / 2

# Done
def get_offdiagonal_of_matrix(matrix: np.ndarray) -> np.ndarray:
    """
    Description:
        Calculates and returns off-diagonal part of given matrix.
    Parameters:
        :param matrix: Matrix for which off-diagonal part shall be calculated.
    Return:
        Off-diagonal part of the matrix.
    """
    matrix_dimension = matrix.shape[0]
    matrix_off_diagonal = copy.copy(matrix)

    for i in range(matrix_dimension):
        # set diagonal element to zero
        matrix_off_diagonal[i, i] = 0

    return matrix_off_diagonal

# Done
def get_offdiagonal_povm_part(povm: List[np.ndarray]) -> List[np.ndarray]:
    """
    Description:
        Get the off-diagonal part of each of POVM's effects.

    Parameters:
        :param povm: POVM - list of matrices representing POVM's effects.

    Return:
        List of np.ndarrays representing off-diagonal parts of POVM's effects.
    """

    # implement function get_off_diagonal_from_matrix for each effect Mi in povm
    return [get_offdiagonal_of_matrix(Mi) for Mi in povm]


def get_diagonal_povm_part(povm: List[np.ndarray]) -> List[np.ndarray]:
    """
    Description:
        From given povm get only diagonal part as  list.

    Parameters:
        :param povm: POVM from effects of which diagonal parts shall be extracted.

    Return:
        List of numpy arrays representing diagonal parts of given POVM.
    """

    # JT: np.diag returns a 1D list of diagonal elements, np.diagflat transforms this into 2D diagonal matrix

    return [np.diagflat(np.diag(effect)) for effect in povm]


def apply_stochastic_map_to_povm(povm: List[np.ndarray],
                                 stochastic_map: np.ndarray) -> List[np.ndarray]:
    number_of_outcomes = len(povm)

    return [sum([stochastic_map[i, j] * povm[j] for j in range(number_of_outcomes)]) for i in range(number_of_outcomes)]


def get_stochastic_map_from_povm(povm: List[np.ndarray]) -> np.ndarray:
    """
    Description:
        From given povm get stochastic map describing classical part of the noise.

    Parameters:
        :param povm: POVM from effects of which diagonal parts shall be extracted.

    Return:
        Stochastic map corresponding to classical part of the noise
    """

    # JT: diagonal elements of a POVM are taken, result is a list of diagonal effects

    diagonal_parts = get_diagonal_povm_part(povm)

    # JT: an empty noise matix is created

    noise_matrix = np.zeros((len(povm), len(povm)), dtype=float)

    for column_index in range(len(povm)):
        # print(diagonal_parts[column_index])

        # JT: a row corresponding to diagonal of a noise matrix is set

        noise_matrix[column_index, :] = np.diag(diagonal_parts[column_index])[:].real

    return noise_matrix


def get_noise_matrix_from_povm(povm):
    """
    See get_stochastic_map_from_povm
    :param povm:
    :type povm:
    :return:
    :rtype:
    """
    # number_of_povm_outcomes = len(povm)
    # dimension = povm[0].shape[0]
    #
    # transition_matrix = np.zeros((number_of_povm_outcomes, number_of_povm_outcomes), dtype=float)
    #
    # for k in range(number_of_povm_outcomes):
    #     current_povm_effect = povm[k]
    #
    #     # Get diagonal part of the effect. Here we remove eventual 0 imaginary part to avoid format conflicts
    #     # (diagonal elements of Hermitian matrices are real).
    #     vec_p = np.array([np.real(current_povm_effect[i, i]) for i in range(dimension)])
    #
    #     # Add vector to transition matrix.
    #     transition_matrix[k, :] = vec_p[:]

    return get_stochastic_map_from_povm(povm=povm)


def get_povm_from_stochastic_map(stochastic_map: np.ndarray) -> List[np.ndarray]:
    return [np.diagflat(stochastic_map[i, :]) for i in range(stochastic_map.shape[0])]


def get_coherent_part_of_the_noise(povm: List[np.ndarray]) -> List[np.ndarray]:
    """
    Description:
        Get the off-diagonal part of each of POVM's effects.

    Parameters:
        :param povm: POVM - list of matrices representing POVM's effects.

    Return:
        List of np.ndarrays representing off-diagonal parts of POVM's effects.
    """
    return get_offdiagonal_povm_part(povm)
#################################################################################################
########## MOVE TO: common.probability #########################################################
####################################################################################################




def find_closest_prob_vector_l2(quasiprobability_vector: List[float]) -> np.ndarray:
    """
    Description:
        Given quasiprobability vector (here it means: vector which elements sum up to 1 but some of them are
        negative), find the closest probability vector in Euclidean norm (2-norm). Here we use fast algorithm from
        Ref. [4] and use the fact that probability distributions are special case of quantum states (namely,
        diagonal quantum states) and fact that for diagonal matrices 2-norm between them is the same as 2-norm
        between vectors created from their diagonal elements.

    Parameters:
        :param quasiprobability_vector: Quasiprobability vector for which closest probability vector will be found.

    Returns:
        Probability vector closest to quasiprobability_vector in Euclidean norm.
    """

    if isinstance(quasiprobability_vector, list):
        d = len(quasiprobability_vector)
    elif isinstance(quasiprobability_vector, type(np.array(0))):
        d = quasiprobability_vector.shape[0]

    # format vector properly
    quasiprobability_vector = np.array(quasiprobability_vector).reshape(d, 1)

    # copy part of the vector
    p000 = list(quasiprobability_vector[:, 0])

    # For algorithm to work we need to rearrange probability vector elements, so we need to keep track of their
    # initial ordering
    p1 = [[i, p000[i]] for i in range(d)]

    # Sort elements in descending order
    p1_sorted = sorted(p1, reverse=True, key=lambda x: x[1])

    # Initiate accumulator
    a = 0

    # go from the i_index=d
    for i in np.arange(0, d)[::-1]:

        # get vector element
        mu_i = p1_sorted[i][1]

        # if mu_i +a/(i_index+1) is negative, do the following:
        if mu_i + a / (i + 1) < 0:
            # add mu_i to accumulator
            a += mu_i

            # set vector element to 0
            p1_sorted[i][1] = 0
        # otherwise
        else:
            # update proper elements of probability vector
            for j in range(i + 1):
                p1_sorted[j][1] += a / (i + 1)
            # finish algorithm - everything is positive now
            break

    # return to initial order
    ordered_p = sorted(p1_sorted, key=lambda x: x[0])

    # get rid of indices
    p_good_format = [ordered_p[i][1] for i in range(d)]

    # print(ordered_p)

    return np.array(p_good_format).reshape(d, 1)


def find_closest_prob_vector_l1(quasiprobability_vector: List[float],
                                method='picos') -> np.ndarray:
    # TODO FBM: perform speed tests between picos/scipy
    """
    Description:
        Given quasiprobability vector (here it means: vector which elements sum up to 1 but some of them are
        negative), find the closest probability vector in l1 norm. Perhaps counterintuively, this
        can be phrased as linear program. See, for example, Ref. [] for nice formulation. #TODO FBM add

    Parameters:
        :param quasiprobability_vector: Quasiprobability vector for which closest probability vector will be found.

    Returns:
        Probability vector closest to quasiprobability_vector in Euclidean norm.
    """

    if isinstance(quasiprobability_vector, list):
        dimension = len(quasiprobability_vector)
    elif isinstance(quasiprobability_vector, type(np.array(0))):
        dimension = quasiprobability_vector.shape[0]

    # format vector properly
    # TODO: we probably don't need it here
    quasiprobability_vector = np.array(quasiprobability_vector).reshape(dimension, 1)

    if method.lower() in ['picos']:
        problem_picos = picos.Problem()

        probability_vector = [picos.RealVariable(name=f'p_{index_outcome}',
                                                 shape=1,
                                                 lower=0,
                                                 upper=1
                                                 )
                              for index_outcome in range(dimension)]
        problem_picos.add_constraint(picos.sum(probability_vector) == 1)

        mus = [picos.RealVariable(name=f'mu_{index_outcome}',
                                  shape=1) for index_outcome in range(dimension)]

        cost_function = 0
        for index_outcome in range(dimension):
            eta_i = quasiprobability_vector[index_outcome] - probability_vector[index_outcome]
            mu_i = mus[index_outcome]

            cost_function += mu_i
            problem_picos.add_constraint(eta_i <= mu_i)
            problem_picos.add_constraint(-eta_i <= mu_i)

            # problem_picos.add_constraint(probability_vector[index_outcome] <= 1)
            # problem_picos.add_constraint(etas[index_outcome] <= mu_i)

        problem_picos.set_objective(direction='min',
                                    expression=cost_function)
        # print(problem_picos)
        problem_picos.solve()

        closest_probability_vector = np.array([problem_picos.get_variable(f'p_{index_outcome}').value
                                               for index_outcome in range(dimension)]).reshape(dimension, 1)
    elif method.lower() in ['scipy']:

        c_vector = np.array([0 for _ in range(dimension)] + [1 for _ in range(dimension)])

        equalities_vector = np.array(1)
        equalities_matrix = np.array([1 for _ in range(dimension)] +
                                     [0 for _ in range(dimension)]).reshape(1, 2 * dimension)

        # inequalities for absolute values
        inequalities_vector = np.zeros((2 * dimension))
        inequalities_vector[0:dimension] = -quasiprobability_vector[:, 0]
        inequalities_vector[dimension:] = quasiprobability_vector[:, 0]

        inequalities_matrix = np.zeros((2 * dimension, 2 * dimension))
        for outcome_index in range(dimension):
            inequalities_matrix[outcome_index, outcome_index] = -1
            inequalities_matrix[outcome_index, outcome_index + dimension] = -1

            inequalities_matrix[outcome_index + dimension, outcome_index] = 1
            inequalities_matrix[outcome_index + dimension, outcome_index + dimension] = -1

        bounds = [(0, 1)
                  # (-quasiprobability_vector[outcome_index],1-quasiprobability_vector[outcome_index])
                  for outcome_index in range(dimension)] + \
                 [(None, None) for _ in range(dimension)]
        print(bounds)
        res = scopt.linprog(c=c_vector,
                            b_eq=equalities_vector,
                            A_eq=equalities_matrix,
                            b_ub=inequalities_vector,
                            A_ub=inequalities_matrix,
                            bounds=bounds,
                            method='interior-point',
                            options={'maxiter': 10 ** 5,
                                     'disp': True}
                            )

        closest_probability_vector = np.array([res.x[outcome_index]
                                               for outcome_index in range(dimension)]).reshape(dimension, 1)

        # print(res.x)

    return closest_probability_vector

# #TODO FBM: FINISH THIS
# dimension_test = 8
# test_quasi = np.array([(-1)**index*np.random.uniform(0.1,1) for index in range(dimension_test)])
# test_quasi/=sum(test_quasi)
#
# print(test_quasi)
#
#
# closest_2_norm = find_closest_prob_vector_l2(quasiprobability_vector=test_quasi)
#
# closest_1_norm = find_closest_prob_vector_l1(quasiprobability_vector=test_quasi)
#
#
# print()
# print('original:',test_quasi)
# print('l2')
# qprint_array(closest_2_norm)
# print('l1')
# qprint_array(closest_1_norm)
# # print(sum(closest_2_norm),sum(closest_1_norm))
#
# print()
# print(np.sqrt(sum([(test_quasi[index_outcome]-closest_2_norm[index_outcome])**2 for index_outcome in range(dimension_test)]))[0])
# print(np.sqrt(sum([(test_quasi[index_outcome]-closest_1_norm[index_outcome])**2 for index_outcome in range(dimension_test)]))[0])
# print(sum([abs(test_quasi[index_outcome]-closest_2_norm[index_outcome]) for index_outcome in range(dimension_test)])[0])
# print(sum([abs(test_quasi[index_outcome]-closest_1_norm[index_outcome]) for index_outcome in range(dimension_test)])[0])
#
#
#
# raise KeyboardInterrupt


def permute_vector(vector, n, transposition):
    # Swap qubits (subspaces) in 2**number_of_qubits dimensional matrix
    # number_of_qubits - number of qubits
    # transposition - which qubits to SWAP.
    # IMPORTANT: in transposition they are numbered from 1

    swap = qubit_swap(n, transposition)
    return swap @ vector


def permute_matrix(matrix, n, transposition):
    # Swap qubits (subspaces) in 2**number_of_qubits dimensional matrix
    # number_of_qubits - number of qubits
    # transposition - which qubits to SWAP.
    # IMPORTANT: in transposition they are numbered from 1

    swap = qubit_swap(n, transposition)
    return swap @ matrix @ swap


def qubit_swap(n, transposition=(1, 1)):
    # create swap between two qubits in 2**number_of_qubits dimensional space
    # labels inside transpositions start from 1.

    D = 2 ** n
    # renumerate for Python convention
    i, j = transposition[0] - 1, transposition[1] - 1

    names = [(bin(j)[2:].zfill(n)) for j in list(range(2 ** n))]

    new_names = copy.copy(names)
    # exchange classical register bits with labels in transposition
    # this defines new order in classical register which respects qubit SWAP
    for k in range(len(new_names)):
        string = list(new_names[k])
        string[i], string[j] = string[j], string[i]
        new_names[k] = ''.join([s for s in string])

    transformation = np.eye(D)

    for x in range(len(names)):
        bit = int(new_names[x], 2)

        # if we need to change the bit, let's do this
        if bit != x:
            transformation[x, x] = 0
            transformation[bit, bit] = 0
            transformation[bit, x] = 1
            transformation[x, bit] = 1

    return transformation


def reorder_classical_register(new_order: Iterable) -> List:
    # reorder classical register according to new_order.
    n = len(new_order)

    # get standard classical register
    standard_register = [(bin(j)[2:].zfill(n)) for j in list(range(2 ** n))]

    return [convert.sort_bitstring(s, new_order) for s in standard_register]




def reorder_probabilities(probabilities, new_order):
    # sort elements of probabilities vector or array according to new_order defined for bits, into a vector

    D = len(probabilities)
    array_format = False
    if isinstance(probabilities, np.ndarray):
        probabilities = probabilities.reshape(D, 1)
        array_format = True

        # get classical register according sorted to new order
    new_classical_register = reorder_classical_register(new_order)
    # sort probabilities entries according to new register
    sorted_probs = utils.sort_things(probabilities, new_classical_register)

    if array_format:
        return np.array(sorted_probs).reshape(D, 1)
    else:
        return sorted_probs


def get_CBT_norm(J, n, m, rev=False):
    import cvxopt as cvx
    import picos as pic
    # Get completely bounded trace norm of Choi-matrix J representing quantum channel from number_of_qubits-dimensional space to m-dimensional space
    J = cvx.matrix(J)
    prob = pic.Problem(verbose=0)
    X = prob.add_variable("X", (n * m, n * m), vtype='complex')

    I = pic.new_param('I', np.eye(m))

    rho0 = prob.add_variable("rho0", (n, n), vtype='hermitian')
    rho1 = prob.add_variable("rho1", (n, n), vtype='hermitian')
    prob.add_constraint(rho0 >> 0)
    prob.add_constraint(rho1 >> 0)

    prob.add_constraint(pic.trace(rho0) == 1)
    prob.add_constraint(pic.trace(rho1) == 1)

    if (rev == True):
        # TODO FBM: tests which conention is good.
        # TODO FBM: add reference to paper

        # This is convention REVERSED with respect to the paper,
        # and seems that this is a proper one????
        C0 = pic.kron(rho0, I)
        C1 = pic.kron(rho1, I)
    else:
        C0 = pic.kron(I, rho0)
        C1 = pic.kron(I, rho1)

    F = pic.trace((J.H) * X) + pic.trace(J * (X.H))

    prob.add_constraint(((C0 & X) // (X.H & C1)) >> 0)

    prob.set_objective('max', F)

    prob.solve(verbose=0)

    if prob.status.count("optimal") > 0:
        #        print('solution optimal')
        1
    elif (prob.status.count("optimal") == 0):
        print('uknown_if_solution_optimal')

    else:
        print('solution not found')

    cbt_norm = prob.obj_value() / 2

    if (abs(np.imag(cbt_norm)) >= 0.00001):
        raise ValueError
    else:
        cbt_norm = np.real(cbt_norm)

    return cbt_norm


def get_choi_from_POVM(POVM):
    # get Choi matrix of POVM channel
    dimension = POVM[0].shape[0]
    number_of_outcomes = len(POVM)
    choi_matrix = np.zeros((number_of_outcomes * dimension, number_of_outcomes * dimension), dtype=complex)

    for index_outcome in range(number_of_outcomes):
        choi_matrix[index_outcome * dimension:
                    (index_outcome + 1) * dimension, index_outcome * dimension:(index_outcome + 1) * dimension] \
            = (POVM[index_outcome].T)[:, :]

    return choi_matrix / dimension


def get_POVM_from_choi(povm_choi,
                       number_of_outcomes,
                       dimension):
    povm = []

    povm_choi *= dimension
    for index_outcome in range(number_of_outcomes):
        block_now = povm_choi[index_outcome * dimension:
                              (index_outcome + 1) * dimension,
                    index_outcome * dimension:(index_outcome + 1) * dimension] \
        #\
        povm.append(block_now.T)
        # = (POVM[index_outcome].T)[:, :]

    return povm




def check_if_povm(povm):
    for Mi in povm:
        eigvals = np.round(sc.linalg.eigvalsh(Mi), 10)

        if any(np.array(eigvals) < 0):
            return False

    return identity_check(sum(povm), significant_digits=10)


def get_statistical_error_bound(number_of_measurement_outcomes: int,
                                number_of_samples: int,
                                statistical_error_mistake_probability: float,
                                number_of_marginals=1) -> float:
    """
    Description:
        Get upper bound for tv-distance of estimated probability distribution from ideal one. See Ref. [3] for
        details.

    Parameters:
        :param number_of_measurement_outcomes: Number of outcomes in probabiility distribution (2^(number_of_qubits) for standard measurement)
        :param number_of_samples: Number of samples for experiment for which statistical error bound is being calculated.
        :param statistical_error_mistake_probability: Parameter describing infidelity of returned error bound.

    Return:
        Statistical error upper bound in total variance distance.
    """

    if number_of_marginals == 0:
        number_of_marginals = 1

    if number_of_measurement_outcomes < 16:
        # for small number of outcomes "-2" factor is not negligible
        return np.sqrt(
            (np.log(2 ** number_of_measurement_outcomes - 2)
             - np.log(statistical_error_mistake_probability) + np.log(
                        number_of_marginals)) / 2 / number_of_samples
        )
        # for high number of outcomes "-2" factor is negligible
    else:
        return np.sqrt(
            (number_of_measurement_outcomes * np.log(2) - np.log(
                statistical_error_mistake_probability) + np.log(
                number_of_marginals)) / 2 / number_of_samples
        )


def get_coherent_error_bound(povm: np.ndarray) -> float:
    """
    Description:
        Get distance between diagonal part of the POVM and the whole POVM. This quantity might be interpreted as a
        measure of "non-classicality" or coherence present in measurement noise. See Ref. [1] for details.

    Parameters:
        :param povm: POVM for which non-classicality will be determined.
    Return:
        Coherent error bound for given POVM.
    """

    return fun_dist.operational_distance_POVMs(povm, get_diagonal_povm_part(povm))

def get_correction_error_bound_from_data_and_statistical_error(povm: List[np.ndarray],
                                                               correction_matrix: np.ndarray,
                                                               statistical_error_bound: float,
                                                               alpha: float = 0) -> float:
    """
        Description:
            Get upper bound for the correction error using classical error-mitigation via "correction matrix".

            Error arises from three factors - non-classical part of the noise, statistical fluctuations and eventual
            unphysical "first-guess" (quasi-)probability vector after the correction.

            This upper bound tells us quantitatively what is the maximal TV-distance of the corrected probability vector
            from the ideal probability distribution that one would have obtained if there were no noise and the
            infinite-size statistics.

            See Ref. [1] for details.

        Parameters:
            :param povm: POVM representing measurement device.
            :param correction_matrix: Correction matrix obtained via out Error Mitigator object.
            :param statistical_error_bound: Statistical error bound (epsilon in paper).
            confidence with which we state the upper bound. See Ref. [3] for details.
            :param alpha: distance between eventual unphysical "first-guess" quasiprobability vector and the closest
            physical one. default is 0 (which corresponds to situation in which corrected vector was proper probability
            vector).


        Return:
            Upper bound for correction error.

        """

    norm_of_correction_matrix = np.linalg.norm(correction_matrix, ord=1)
    coherent_error_bound = get_coherent_error_bound(povm)
    return norm_of_correction_matrix * (coherent_error_bound + statistical_error_bound) + alpha


def get_correction_error_bound_from_data(povm: List[np.ndarray],
                                         correction_matrix: np.ndarray,
                                         number_of_samples: int,
                                         error_probability: float,
                                         alpha: float = 0) -> float:
    """
    Description:
        Get upper bound for the correction error using classical error-mitigation via "correction matrix".

        Error arises from three factors - non-classical part of the noise, statistical fluctuations and eventual
        unphysical "first-guess" (quasi-)probability vector after the correction.

        This upper bound tells us quantitatively what is the maximal TV-distance of the corrected probability vector
        from the ideal probability distribution that one would have obtained if there were no noise and the
        infinite-size statistics.

        See Ref. [0] for details.

    Parameters:
        :param povm: POVM representing measurement device.
        :param correction_matrix: Correction matrix obtained via out Error Mitigator object.
        :param number_of_samples: number of samples (in qiskit language number of "shots").
        :param error_probability: probability with which statistical upper bound is not correct. In other word, 1-mu is
        confidence with which we state the upper bound. See Ref. [3] for details.
        :param alpha: distance between eventual unphysical "first-guess" quasiprobability vector and the closest
        physical one. default is 0 (which corresponds to situation in which corrected vector was proper probability
        vector).


    Return:
        Upper bound for correction error.

    """
    dimension = povm[0].shape[0]

    norm_of_correction_matrix = np.linalg.norm(correction_matrix, ord=1)

    statistical_error_bound = get_statistical_error_bound(dimension, number_of_samples,
                                                          error_probability)
    coherent_error_bound = get_coherent_error_bound(povm)

    return norm_of_correction_matrix * (coherent_error_bound + statistical_error_bound) + alpha


def get_correction_error_bound_from_parameters(norm_of_correction_matrix: float,
                                               statistical_error_bound: float,
                                               coherent_error_bound: float,
                                               alpha: float = 0) -> float:
    """
    Description:
        See description of function "get_correction_error_bound_from_data". This function can be used if one has the
        proper parameters already calculated and wishes to not repeat it (for example, in case of calculating something
        in the loop).

    Parameters:
        :param norm_of_correction_matrix : 1->1 norm of correction matrix (it is not trace norm!), see Ref. [0],
        or definition of np.linalg.norm(X,ord=1)
        :param statistical_error_bound: upper bound for statistical errors. Can be calculated using function
        get_statistical_error_bound.
        :param coherent_error_bound: magnitude of coherent part of the noise. Can be calculated using function
        get_coherent_error_bound.
        :param alpha: distance between eventual unphysical "first-guess" quasi-probability vector and the closest
        physical one. default is 0 (which corresponds to situation in which corrected vector was proper probability
        vector)

    Return:
        Upper bound for correction error.
    """

    return norm_of_correction_matrix * (coherent_error_bound + statistical_error_bound) + alpha


# ORGANIZE: This manipulates results data and should probably fit better in functions_data_analysis
def counts_dict_to_frequencies_vector(count_dict: dict, reverse_order=False) -> list:
    """
    Description:
        Generates and returns vector of frequencies basing on given counts dict. Mostly used with qiskit data.
    :param count_dict: Counts dict. Possibly from qiskit job.
    :return frequencies: Frequencies list for possible states in ascending order.
    """

    frequencies = []

    qubits_number = len(list(count_dict.keys())[0])  # Number of qubits in given experiment counts.
    possible_outcomes = get_possible_n_qubit_outcomes(qubits_number)
    dict_keys = count_dict.keys()  # So we don't call the method_name every time.
    counts_sum = 0

    for outcome in possible_outcomes:
        if dict_keys.__contains__(outcome):
            frequencies.append(count_dict[outcome])
            counts_sum += count_dict[outcome]
        else:
            frequencies.append(0)

    for i in range(len(frequencies)):
        frequencies[i] = frequencies[i] / counts_sum

    if reverse_order:
        return reorder_probabilities(frequencies, range(qubits_number)[::-1])
    else:
        return frequencies


def get_possible_n_qubit_outcomes(n: int) -> list:
    """
    Description:
        For given number of qubits <n> generates a list of possible outcome states (as strings) and returns them in
        ascending order. All states len is number_of_qubits.
    :param n: Number of qubits.
    :return: List of possible outcomes as strings.
    """
    max_value = pow(2, n)
    possible_outcomes = []

    for i in range(max_value):
        possible_outcomes.append(bin(i)[2:].zfill(n))

    return possible_outcomes


def get_correction_matrix_from_povm(povm):
    number_of_povm_outcomes = len(povm)
    dimension = povm[0].shape[0]

    transition_matrix = np.zeros((number_of_povm_outcomes, number_of_povm_outcomes), dtype=float)

    for k in range(number_of_povm_outcomes):
        current_povm_effect = povm[k]

        # Get diagonal part of the effect. Here we remove eventual 0 imaginary part to avoid format conflicts
        # (diagonal elements of Hermitian matrices are real).
        vec_p = np.array([np.real(current_povm_effect[i, i]) for i in range(dimension)])

        # Add vector to transition matrix.
        transition_matrix[k, :] = vec_p[:]

    return np.linalg.inv(transition_matrix)


def get_averaged_lambda(big_lambda,
                        bits_of_interest):
    big_N = int(np.log2(big_lambda.shape[0]))
    small_N = len(bits_of_interest)
    normalization = 2 ** (big_N - small_N)

    classical_register_big = ["{0:b}".format(i).zfill(big_N) for i in range(2 ** big_N)]
    classical_register_small = ["{0:b}".format(i).zfill(small_N) for i in range(2 ** small_N)]

    indices_small = {}

    for s in classical_register_big:
        small_string = ''.join([list(s)[b] for b in bits_of_interest])
        indices_small[s] = small_string

    small_lambda = np.zeros((2 ** small_N, 2 ** small_N))

    for i in range(2 ** big_N):
        for j in range(2 ** big_N):
            lambda_element = big_lambda[i, j]
            ideal_state = "{0:b}".format(j).zfill(big_N)
            measured_state = "{0:b}".format(i).zfill(big_N)

            ideal_state_small = indices_small[ideal_state]
            measured_state_small = indices_small[measured_state]

            small_lambda[int(measured_state_small, 2), int(ideal_state_small,
                                                           2)] += lambda_element / normalization
    return small_lambda



# REMOVE:
# ---------------------------------------------------------------
# def get_tensored_POVM(POVMS_list):
#     from qrem.functions_qrem.PyMaLi import \
#         GeneralTensorCalculator  # old version: from PyMaLi import GeneralTensorCalculator
#     gtc = GeneralTensorCalculator()
#     return gtc.calculate_tensor_to_increasing_list(objects=POVMS_list)



"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""
# DELETE THIS FILE -- talk with Jan what is inside (not sure if this is a real test)
# This test uses specific data which is test data - should be moved maybe to test folder?
#check if this test is 

import pickle

from qrem.functions_qrem import ancillary_functions as anf, quantum_ancillary_functions as quanf
from qrem.common import io

# specify data used for testing
backend_name = 'ASPEN-9'

date, bitstrings_right_to_left, number_of_qubits = anf.get_historical_experiments_number_of_qubits(
    backend_name=backend_name)

date = '2021-09-27'

module_directory = anf.get_local_storage_directory()

tests_directory_main = module_directory + '/saved_data/characterization2021/'
tests_directory_low = f"{date}/{backend_name}/N{number_of_qubits}/"
directory_to_open = tests_directory_main + 'processed_data/' + tests_directory_low

tests_directory_low = f"ground_states_implementation/2SAT/{date}/N{number_of_qubits}/"
directory_to_open = tests_directory_main + 'processed_data/' + tests_directory_low

with open(directory_to_open + "00_raw_results.pkl", 'rb') as filein:
    experiments_results = pickle.load(filein)

with open(directory_to_open + "full_information_2SAT.pkl", 'rb') as filein:
    hamiltonians_data = pickle.load(filein)

# io.save(dictionary_to_save=hamiltonians_data['N23'],
#                         directory=directory_to_open,
#                         custom_filename="full_information_2SAT_N23")

target_number_of_qubits = 20
better_hamiltonians_data = {}


for hamiltonian_index, ham_dictionary in hamiltonians_data['N20'].items():

    # raise KeyError
    weights_dictionary = ham_dictionary['weights_dictionary']
    ground_state = ham_dictionary['ground_state']

    weights_dictionary_better = {anf.convert_qubits_string_to_tuple(qubits_string=qubits_string):weight
                          for qubits_string,weight in weights_dictionary.items()}

    ideal_energy = quanf.get_energy_from_bitstring_diagonal(bitstring=ground_state,
                                                            weights_dict=weights_dictionary_better)
    ham_dictionary['ground_state_energy'] = ideal_energy
    ham_dictionary['weights_dictionary'] = weights_dictionary_better



    fixed_qubits = list(range(target_number_of_qubits,len(ground_state)))
    ground_state_of_fixed_qubits = [ground_state[x] for x in fixed_qubits]

    string_ground_state_of_fixed_qubits = ''.join(ground_state_of_fixed_qubits)

    ground_state_of_variable_qubits = [ground_state[x] for x in list(range(target_number_of_qubits))]
    string_ground_state_now = ''.join(ground_state_of_variable_qubits)

    # print(ground_state)
    # print(string_ground_state_now)
    # print(fixed_qubits)
    # print(string_ground_state_of_fixed_qubits)
    # raise KeyError

    if string_ground_state_now in better_hamiltonians_data.keys():
        better_hamiltonians_data[string_ground_state_now][string_ground_state_of_fixed_qubits] = ham_dictionary
    else:
        better_hamiltonians_data[string_ground_state_now] = {string_ground_state_of_fixed_qubits: ham_dictionary}



    #
print('saving')
io.save(dictionary_to_save=better_hamiltonians_data,
                        directory=directory_to_open,
                        custom_filename="full_information_2SAT_N20")

    # raise KeyError
#     print(number_of_qubits)



#MOcomm - do we need and use this file??
#ORGANIZE - delate/ move?
import pickle

from tqdm import tqdm
import numpy as np

from qrem.functions_qrem import ancillary_functions as anf
from qrem.functions_qrem import functions_data_analysis as fda
from qrem.functions_qrem import functions_standarized_directories as dirs

from qrem.noise_simulation.CN import sampling_improvements as si

from qrem.common import io
from qrem.common.printer import qprint

# specify data used for testing
backend_name = 'ASPEN-9'
date = '2021-09-27'
number_of_qubits = 20
bitstrings_right_to_left = False


backend_name = 'ibm_washington'
date = '2022-04-10'
number_of_qubits = 109



# Specify whether save calculated data
saving = True

running_estimation = False


noise_model_name = "uncorrelated"

noise_model_name = "only_clusters"

alpha_multiplier = 1.0

if noise_model_name.upper()=='ONLY_CLUSTERS':
    no_neighbors = True
    noise_model_name = f"{noise_model_name}_alpha-{alpha_multiplier}"



directory_merged_noise_models_SCR= dirs.get_directory_processed_experimental_results(backend_name=backend_name,
                                                                                                  date=date,
                                                                                                  number_of_qubits=number_of_qubits,
                                                                                                  experiment_name='noise_mitigation',
                                                                                                  additional_path="noise_models_data/merged_models/SCR/")

folder_name_hamiltonians = 'ground_states_implementation/2SAT/'


directory_stored_hamiltonian = dirs.get_directory_stored_hamiltonians(number_of_qubits=number_of_qubits,
                                                                      hamiltonian_name='2SAT')
with open(directory_stored_hamiltonian + "full_information_2SAT_N20.pkl", 'rb') as filein:
    hamiltonians_data = pickle.load(filein)

hamiltonian_weights_all = {input_state:
                           ham_data['']['weights_dictionary']
                           for input_state, ham_data in hamiltonians_data.items()}


directory_noise_mitigation= dirs.get_directory_processed_experimental_results(backend_name=backend_name,
                                                                                      date=date,
                                                                                      number_of_qubits=number_of_qubits,
                                                                                      experiment_name='noise_mitigation')



directory_noise_characterization = dirs.get_directory_processed_experimental_results(backend_name=backend_name,
                                                                                        date=date,
                                                                                        number_of_qubits=number_of_qubits,
                                                                                        experiment_name='noise_characterization')

directory_energy_estimation_results = dirs.get_directory_processed_experimental_results(backend_name=backend_name,
                                                                                        date=date,
                                                                                        number_of_qubits=number_of_qubits,
                                                                                        experiment_name='ground_states_implementation/2SAT')
with open(directory_energy_estimation_results + "energy_estimations_counts.pkl", 'rb') as filein:
    experiments_results = pickle.load(filein)

results_dictionary = experiments_results['results_dictionary']

experiment_keys = list(results_dictionary.keys())

with open(directory_energy_estimation_results + 'estimated_energies_results.pkl', 'rb') as filein:
    estimated_energies = pickle.load(filein)

with open(directory_noise_mitigation + f'correction_data_merged_{noise_model_name}.pkl', 'rb') as filein:
    dictionary_mitigation_data = pickle.load(filein)


dictionary_results_noise_matrices = dictionary_mitigation_data['joint_noise_matrices_dictionary']



joint_correction_matrices_dictionary = dictionary_mitigation_data[
    'joint_correction_matrices_dictionary']
# noise_models_correction_indices = dictionary_mitigation_data['correction_indices']
joint_noise_matrices_dictionary = dictionary_mitigation_data['joint_noise_matrices_dictionary']

# raise KeyboardInterrupt
do_sanity_checks = False

test_cl = [(x,) for x in range(number_of_qubits)]

all_pairs = [(i, j) for i in range(number_of_qubits) for j in range(i + 1, number_of_qubits)]

with open(directory_merged_noise_models_SCR + f'noise_models_merged_data_SCR_{noise_model_name}.pkl', 'rb') as filein:
    dictionary_data_noise_models = pickle.load(filein)


number_of_samples = 10**6

#TODO FBM: fix those indices, something seems to be wrong.

with open(directory_noise_characterization + 'noise_model_identifiers2.pkl', 'rb') as filein:
    noise_model_identifiers = pickle.load(filein)


run_index = 0
qprint("\nRUN INDEX:",run_index,'red')
range_run= slice(run_index*2,(run_index+1)*2)
print(range_run)





# print(len(list(noise_model_identifiers.keys())))
# print(len(list(dictionary_data_noise_models.items())))
#
# print(list(noise_model_identifiers.keys())[0])
for clusters_key, noise_model_data in list(dictionary_data_noise_models.items())[range_run]:
    clusters_list = list(clusters_key)
    local_noise_matrices = noise_model_data['local_noise_matrices']

    noise_parameters = noise_model_data['noise_model_parameters']

    noise_parameters = sorted(noise_parameters, key=lambda x:x[0]+x[2])
    noise_parameters = tuple(noise_parameters)
    # print(noise_parameters)
    #
    # raise KeyboardInterrupt

    noise_model_identifier_now = noise_model_identifiers[noise_parameters]


    qprint("\nCurrent noise model", clusters_list)
    print(noise_model_data.keys())

    # SCR_dict_now = noise_model_data['SCR_dictionary']

    symmetrizers_dictionary = noise_model_data['symmetrizers_dictionary']
    error_probabilities_dictionary = noise_model_data['error_probabilitities_dictionary']

    print(error_probabilities_dictionary)

    qubits_to_clusters_map = {}
    for qubit in range(number_of_qubits):
        for cluster in clusters_list:
            if qubit in cluster:
                qubits_to_clusters_map[qubit] = cluster



    multipliers_mitigation = {}
    for i in range(number_of_qubits):
        for j in range(i+1,number_of_qubits):
            multiplier = 1
            for qi in (i,j):
                multiplier/=(1 - 2 * error_probabilities_dictionary[(qi,)])
            multipliers_mitigation[(i,j)] = multiplier
    for i in range(number_of_qubits):
        multipliers_mitigation[(i,)] = 1/(1 - 2 * error_probabilities_dictionary[(i,)])


    results_dictionary_models = {}
    figures_of_merit_list = []
    for input_state in tqdm(experiment_keys):
        energy_estimated = estimated_energies[input_state]
        hamiltonian_dict = hamiltonians_data[input_state]['']
        weights_dictionary = hamiltonian_dict['weights_dictionary']
        needed_marginals = list(weights_dictionary.keys())
        energy_ideal = hamiltonian_dict['ground_state_energy']


        results_now_dict = results_dictionary[input_state]
        unique_counts = list(results_now_dict.keys())
        probability_distro = np.array(list(results_now_dict.values()),dtype=float)
        numer_of_shots_in_experiments = sum(probability_distro)
        probability_distro/=numer_of_shots_in_experiments
        if number_of_samples==sum(list(results_now_dict.values())):
            resampled_outcomes_dictionary = results_now_dict
        else:
            resampled_outcomes = np.random.multinomial(pvals=probability_distro,
                                                       n=number_of_samples)


            resampled_outcomes_dictionary = {unique_counts[outcome_index]:resampled_outcomes[outcome_index]
                                             for outcome_index in range(len(resampled_outcomes))}


        equalized_samples = si.sample_from_multiple_input_states_alternative(input_states_dictionary=resampled_outcomes_dictionary,
                                                                             local_stochastic_matrices=symmetrizers_dictionary,
                                                                             print_progres_bar=False,
                                                                             # old_sampling=True
                                                                 )

        # energy_corrected = fda.estimate_energy_from_counts_dictionary(counts_dictionary=equalized_samples,
        #                                                               weights_dictionary=weights_dictionary,
        #                                                               additional_multipliers=multipliers_mitigation)
        # print(equalized_samples)
        energy_corrected = fda.estimate_energy_from_counts_dictionary_alternative(counts_dictionary=equalized_samples,
                                                                      weights_dictionary=weights_dictionary,
                                                                      additional_multipliers=multipliers_mitigation,
                                                                                  )

        # t3 = time.time()

        # qprint("TOTAL ENERGY ESTIMATION TIME:",t3-t2)

        error_corrected = abs(energy_corrected-energy_ideal)/number_of_qubits
        error_estimated = abs(energy_estimated-energy_ideal)/number_of_qubits
        figure_of_merit = error_corrected/error_estimated
        figures_of_merit_list.append(figure_of_merit)
        # print(energy_corrected==energy_corrected1)
        # print(energy_corrected-energy_corrected2)

        #
        # qprint("Ratio newmethod/oldmethod:",(dt0/dt1))
        # qprint("number of samples",number_of_samples)



        # raise KeyboardInterrupt

        # qprint("Energy ideal:",energy_ideal)
        qprint("\nEnergy estimated error per qubit:",np.round(error_estimated,3))
        # qprint("Energy corrected:", energy_corrected)
        qprint("Energy corrected error per qubit:", np.round(error_corrected,3))
        qprint("Figure of merit", np.round(figure_of_merit,3))
        print()

        # print(energy_ideal,
        #       energy_estimated,
        #       energy_corrected,
        #       abs(energy_corrected - energy_ideal) / number_of_qubits
        #       # en_noisy
        #       )

        # raise KeyboardInterrupt


        results_dictionary_models[input_state] = {
            'error_corrected': error_corrected,
            'error_estimated': error_estimated,
            'energy_corrected': energy_corrected,
            'energy_estimated': energy_estimated,
            'energy_ideal': energy_ideal,

        }

    qprint("Mean figure of merit:", np.mean(figures_of_merit_list))
    io.save(dictionary_to_save=results_dictionary_models,
                            directory=directory_energy_estimation_results + f'/noise_mitigated/SCR/particular_models/s_{number_of_samples}/',
                            custom_filename="id_" + str(noise_model_identifier_now))





from typing import Dict, Tuple, List

"""x, y - bit strings
j, k - indices

Returns True if x=y outside indices {j, k} and False otherwise.
"""

def outeq(x: str, y: str, j: int, k:int) -> bool:
    # todo: what if n=2 ?
    return x[:j] + x[j+1:k] + x[k+1:] == y[:j] + y[j+1:k] + y[k+1:]
